package cs297.xmldb.src.bm;

import java.io.*;
import java.util.*;
import java.lang.*;
import cs297.xmldb.src.cm.*;
import cs297.xmldb.src.sm.*;
import cs297.xmldb.src.server.*;

/**
 @Date: Spring & fall 2004
 @author: Thien An Nguyen, 
 A data page with additional operations defined from NATIX's algorithm.
 */
public class ThPageInterpreter extends ThPage implements IThPage
{  
   ArrayList localLogs;
   Hashtable localTM;
   String previousTX = ""; 
   String lastIn = "";
   boolean preState = false;
  
   public ThPageInterpreter(int size, int pnum)
   {
      super(size, pnum);
      localLogs = new ArrayList();
      localTM = new Hashtable();
   }
  
   public ThPageInterpreter(int size, byte[] bytes)
   {
      super(size, bytes);
      localLogs = new ArrayList();
      localTM = new Hashtable();
   }
  
   public char[] toHex() 
   {
      return super.toHex();
   }
  
   public int getPageNum()
   {
      return super.getPageNum();
   }
  
   /**
   Return a last LSN of the requested transaction id
   @param transaction id
   @returns: the last LSN of requested tx id.
   */
   public String getLSN(String txid)
   {
      ThTransaction tx = (ThTransaction) localTM.get(txid);
      ArrayList lsns = new ArrayList();
      
      if (tx != null)
      {
         tx.getLSNsForRollback(lsns);
         return (String) lsns.get(lsns.size()-1);
      }
      return null;
   }
  
   /**
   Get the log record given its lsn
   @param The LSN of the requested log
   @returns: The log.
   */
   public ThLogRecord getLog(String lsn)
   {
      ThLogRecord log;
      for (int i=localLogs.size()-1; i>=0; i--)
      {
         log = (ThLogRecord) localLogs.get(i);
         if (lsn.equals(log.getLSN()))
         return log;
      }
    
      return null;
   }
  
   /**
   Overide the base's createInsertLogRecord to format logs according to NATIX.
   */
   public String createInsertLogRecord(String rectype, String txid,
            int id, String path, ArrayList values,  String xmlname, 
            String dbname, ThSystem sys, boolean natixlog)
   {
      String lsn = "";
      if (previousTX.equals(txid))
      {
         if (!preState && rectype.equals("UINSERTLEAVE") 
          && lastIn.equals("G"))
         {
            preState = true;
         }
         else if (preState && rectype.equals("UINSERTLEAVE")
           && lastIn.equals("L"))
         {
            preState = true;
         } 
         else if (preState && rectype.equals("UINSERTGROUP")
           && lastIn.equals("L"))
         {
            preState = false;
         }
         else if (!preState && rectype.equals("UINSERTGROUP")
           && lastIn.equals("G"))
         {
            preState = false;
         }
           
         lastIn = rectype.substring(7,8);
         if (preState)
         {
            lsn = getLSN(txid);
            if (lsn != null) 
            {
            //get the previous log record out and add into that
            ThLogRecord log = getLog(lsn);
            if (log != null)
            {
               log.addLeave((String) values.get(0));
               return lsn;
            }
            }
         }
      } 
         lastIn = rectype.substring(7,8);
         publicLogsToLM(sys);
         lsn = sys.getLSN();      
         ThLogRecord log = new ThLogRecord(rectype,lsn,
                        Integer.toString(getPageNum()),
                        txid, Integer.toString(id), path, values, 
                        xmlname, dbname);
        
         localLogs.add(log); 
         addUndoLSN(txid, lsn);
         previousTX = txid;
      
      return lsn;
   }
  
   public String createDelLogRecord(String rectype, String txid, 
                                int id, String path, 
                                String xmlname, String dbname, ThSystem sys)
   {
      return super.createDelLogRecord(rectype, txid, id, path,
                                   xmlname, dbname, sys);
   }  
  
   public String createUpdateLogRecord(String rectype, String txid,
               int id, String path, String oldvalue, String newvalue,
               String xmlname, String dbname, ThSystem sys)
   {
      return super.createUpdateLogRecord(rectype, txid, id,
              path, oldvalue, newvalue, xmlname, dbname, sys);
   } 
  
   /**
   Over write the base's addUndoLSN
   @param Transactionid,LSN of the transaction operation.
   @returns: Return Code.
   */
   public int addUndoLSN(String txid, String lsn)
   {
      if (!localTM.containsKey(txid))
        localTM.put(txid, new ThTransaction(txid));
    
      ((ThTransaction)localTM.get(txid)).addUndoLSN(lsn);
    
      return 0;
   }
  
   /**
   Over write the ThPage's insert method.
   */
   public int insert(String msg, String txid, int id, String path,
                    ArrayList value, String xmlname, String dbname,
                    boolean isgroup, ThSystem sys, boolean natixlog,
                    boolean islog)
   {
      super.inFo.set(7); // set dirty bit on
      if (islog)
      {
         String lsn = createInsertLogRecord(msg, txid, id, path, 
                           value, xmlname, dbname, sys, natixlog);
                                  
         super.pageLSN = Integer.parseInt(lsn); // update pageLSN
      }
      
      ThNode node = new ThNode(id, isgroup, value);
      super.nodeS.add(node);
      
      return 0; //success
   }  

   public int insert(IThNode node, int parentid, ThSystem sys) 
   {
      return -1;
   }
    
   public int insert(IThNode newstructednode, IThNode prev, ThSystem sys)
   {
     return -1;
   }
  
   public int delete(int id)
   {
      return super.delete(id);
   }
  
   public String update(int id, String value, String lsn, ThSystem sys) 
   {
      return super.update(id, value, lsn, sys);
   }
  
   public void updateFreeSpace(int reclen)
   {
      super.updateFreeSpace(reclen);
   }
      
   public ArrayList getData(int id) 
   {
      return super.getData(id);
   }
 
   public boolean isFull(int length) 
   {
      return super.isFull(length);
   }
     
   public int getSize()
   {
      return super.getSize();
   }
  
   public int getOffset(int id)
   {
      return super.getOffset(id);
   }
  
   public ThNode getDataNode(int id)
   {    
      return super.getDataNode(id);
   }
         
   public IThNode remove(int id) 
   {
      return super.remove(id);
   }
  
   public int getNodeCount() 
   {
      return super.getNodeCount();
   }
  
   public IThNode getNodeByOffset(int idx) 
   {
      return super.getNodeByOffset(idx);
   }  
  
   /**
   Public local log records to Log Manager
   @param ThSystem where global variables held
   @returns: nothing
   */
   public void publicLogsToLM(ThSystem sys) 
   {        
      if (localLogs.size() < 1)
         return;
      //go through the log records, call LM to write to disk
      ThLogRecord log;
      String logdata;
      String logtype, txid, lsn;
      ArrayList values;
    
      while (localLogs.size()>0) 
      {
         log = (ThLogRecord) localLogs.get(0);
         logtype = log.getLogType();
         txid = log.getTxId();
         lsn = log.getLSN();
         logdata =  logtype+ " "+ lsn + " "
                  + log.getDbName()+ " "+ log.getXmlName() + " "
                  + log.getPageNum() + " " 
                  + txid   + " " + log.getId()+" "
                  + log.getPath()   + " " + "(";  
       
        values = log.getNewValues();
        logdata += (String) values.get(0); 
        if (logtype.equals("UINSERTGROUP"))
        {
           logdata += " "+(String) values.get(1); 
         
           for (int j=2; j<values.size(); j++)
           logdata += ") ("+(String) values.get(j);   
        }
        logdata += ")";   
       
        sys.getLM().writeLogRecord(logdata);    
        sys.getTM().addUndoLSN(txid, lsn);
       
        localLogs.remove(0);
     } 
     localTM = new Hashtable(); 
     resetStates();
   }
    
   public ThLogRecord getLogRecord(String lsn)
   {
      ThLogRecord log;
      for (int i=0; i<localLogs.size();i++)
      {
         log = (ThLogRecord) localLogs.get(i);
         if (lsn.equals(log.getLSN()))
         return log;
      }
    
      return null;
   }  
   
   private void resetStates()
   {
      preState = false;
   }
   
   public void add(IThNode node) {
   }
   
   public void removeSubTree(ThGroup node) {
   }
   
   public void addSubTree(ThGroup node) {
   }
}