import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.net.*;

/**
 * @author: Thien An Nguyen, 
 */
public class ThLoginDialog extends JDialog implements ActionListener
{
	//input fields
	private 	JTextField octet1,
					 octet2,
					 octet3,
					 octet4;
	private JTextField userName;
	private JPasswordField	passWord;
	private JButton loginButton;
	private JButton closeButton;
	//connection info.
	private String uId;
	private String psswd;
	private String host;
	private boolean abort = false;
	ObjectInputStream in;
	ObjectOutputStream out;
	//address info.
	Socket socket = null;

	/**
	* constructs graphic user interface components
	*/
	public ThLoginDialog(JFrame parent)
	{
		super(parent, "Server name and security check", true);
		addWindowListener(new WindowAdapter(){  
			public void windowClosing(WindowEvent e){
				abort = true;
				System.exit(0);
			}
		 } );
		
		Container contentPane = getContentPane();
		JPanel p1 = new JPanel();

	//set up lay out attributes
		GridBagLayout layout = new GridBagLayout();
		p1.setLayout(layout);
		GridBagConstraints c = new GridBagConstraints();
		c.weightx = 100;
		c.weighty = 100;                   
	  	c.gridheight = 1;  

	//user name
		JLabel l1 = new JLabel("Host Name: ");
	 	c.gridwidth = 10;
	 	c.anchor = GridBagConstraints.EAST;
		c.gridx = 0;                     
	 	c.gridy = 0;
	  	layout.setConstraints(l1, c);          
		p1.add(l1);

		JLabel l2 = new JLabel("User Name: ");
		c.gridy = 1;
	  	layout.setConstraints(l2, c);          
		p1.add(l2);

		JLabel l3 = new JLabel("Password: ");
		c.gridy = 2;
	  	layout.setConstraints(l3, c);          
		p1.add(l3);

	//textfields
		octet1 = new JTextField(3);
		octet1.setText("127");
		c.gridwidth = 4;
 		c.anchor = GridBagConstraints.CENTER;
		c.fill = GridBagConstraints.HORIZONTAL;			
		c.gridx = 10;                     
	 	c.gridy = 0;
	 	layout.setConstraints(octet1, c);
	 	p1.add(octet1);

	 	octet2 = new JTextField(3);
	 	octet2.setText("0");
		c.gridx = 15;
		layout.setConstraints(octet2, c);	
		p1.add(octet2);

		octet3 = new JTextField(3);
		octet3.setText("0");
		c.gridx = 20;
		layout.setConstraints(octet3, c);
		p1.add(octet3);

		octet4 = new JTextField(3);
		octet4.setText("1");
		c.gridx = 25;	
		layout.setConstraints(octet4, c);
		p1.add(octet4);

		JLabel d2 = new JLabel(".");
		c.gridwidth = 1;
		c.fill = GridBagConstraints.NONE;
		c.gridx = 14;
	  	layout.setConstraints(d2, c);          
		p1.add(d2);		

		JLabel d1 = new JLabel(".");
		c.gridx = 19;
	  	layout.setConstraints(d1, c);          
		p1.add(d1);		

		JLabel d3 = new JLabel(".");
		c.gridx = 24;
	  	layout.setConstraints(d3, c);          
		p1.add(d3);			

		userName = new JTextField(8);
		userName.setText("project1");
		c.gridwidth = 19;
		c.fill = GridBagConstraints.HORIZONTAL;	
		c.anchor = GridBagConstraints.WEST;	
		c.gridx = 10;
		c.gridy = 1;
	  	layout.setConstraints(userName, c);          
		p1.add(userName);
						
		passWord = new JPasswordField(8);
		passWord.setEchoChar('*');
		c.gridy = 2;
	  	layout.setConstraints(passWord, c);          
		p1.add(passWord);

	//log in & close button
		loginButton = new JButton("Login");
		c.gridwidth = 10;
		c.fill = GridBagConstraints.NONE;
		c.anchor = GridBagConstraints.CENTER;	
		c.gridx = 9;
	 	c.gridy = 3; 	 	
		loginButton.addActionListener(this);
	  	layout.setConstraints(loginButton, c);          
		p1.add(loginButton);

		closeButton = new JButton("Close");
		c.gridx = 15; 
		closeButton.addActionListener(this);
	  	layout.setConstraints(closeButton, c);          
		p1.add(closeButton);
		
	//add in the dialog
		contentPane.add(p1);
		setSize(250, 150);
		show();
	}
	/**
	 * when user click on "Login" button, check username and password. 
	 * If "valid", proceed. If "invalid", keep display invalid dialog box until user decide to "close"
	 */
	public void actionPerformed(ActionEvent evt)
	{
		try {
			Object source = evt.getSource();
			
			if(source == loginButton){				
				if (handleLogin()){
					setVisible(false);
				}
			}else if (source == closeButton){
				//dispose();
				abort = true;
				System.exit(0);
			}
		} catch(Exception e) {
			System.err.println("in ThLoginDialog.actionPerformed" + e);
		}
	}
	/**
	* return the host 4 octet number
	*/
	public String getHost()
	{
		return host;
	}
	/**
	* returns the input stream connection
	*/
	public ObjectInputStream getInputObject()
	{
		return in;
	}
	/**
	* returns output stream connection
	*/
	public ObjectOutputStream getOutputObject()
	{
		return out;
	}
	/**
	* returns user password
	*/
	public String getPassword()
	{
		return psswd;
	}
	/**
	* returns the socket
	*/
	public Socket getSocket()
	{
		return socket;
	}
	/**
	* returns user name
	*/
	public String getUserName()
	{
		return uId;
	}
	/**
	* returns true if no error. Returns false if an error occur.
	*/
	public boolean handleLogin()
	{
		return preCheckHost() && preCheckId() && preCheckPssWd() && isTimeServer();
	}
	/**
	* returns true if user clicked close. 
	* returns false otherwise.
	*/
	public boolean isAbort()
	{
		return abort;
	}
	/**
	* returns true if the host entered by user is verified.
	* returns false if error occurs
	*/
	public boolean isTimeServer()
	{
		String received;
		String request = ThTime.TEST + "-" + "Username" + uId + "-" +"Password"+psswd;
		Object response;

		try{
			socket = SocketOpener.openSocket(host, ThTime.PORT, ThTime.TIME_OUT);
			if(socket == null){
				JOptionPane.showMessageDialog(null, "Please check your host and or username/password again!",
									 					"Error!", JOptionPane.ERROR_MESSAGE);
			}else{			
				out = new ObjectOutputStream(socket.getOutputStream());
				out.writeObject(request);

				in = new ObjectInputStream(socket.getInputStream());
				response = (String)in.readObject();
			
				if (response.equals(ThTime.CHECKED)){
					return true;
				}else if (response.equals(ThTime.FAILED)){
					JOptionPane.showMessageDialog(null, "Please check your password again!",
									 					"Error!", JOptionPane.ERROR_MESSAGE);
					return false;
				}
			}
		}catch (Exception e){
			JOptionPane.showMessageDialog(null, "Please check your host and or username/password again!", 
				 					"Error!", JOptionPane.ERROR_MESSAGE);
		}
		
		return false;
	}
	/**
	* returns true if 4 octets fields's length are no greater than 3 digits
	* returns false otherwise.
	*/
	public boolean preCheckHost()
	{
		if ((octet1.getText()).length() > 3){
			JOptionPane.showMessageDialog(null, "First octet should be no more than 3 digits!", 
				 					"Error!", JOptionPane.ERROR_MESSAGE);
			return false;
		}
		if ((octet2.getText()).length() > 3){
			JOptionPane.showMessageDialog(null, "Second octet should be no more than 3 digits!", 
				 					"Error!", JOptionPane.ERROR_MESSAGE);
			return false;
		}
		if ((octet3.getText()).length() > 3){
			JOptionPane.showMessageDialog(null, "Third octet should be no more than 3 digits!", 
				 					"Error!", JOptionPane.ERROR_MESSAGE);
			return false;
		}
		if ((octet4.getText()).length() > 3){
			JOptionPane.showMessageDialog(null, "Fourth octet should be no more than 3 digits!", 
				 					"Error!", JOptionPane.ERROR_MESSAGE);
			return false;
		}

		host = octet1.getText() + "." + octet2.getText() + "." + octet3.getText() + "." + octet4.getText();	
		
		return true;
	}
	/**
	* returns true if username's length is exactly 8 characters.
	* returns false otherwise.
	*/
	public boolean preCheckId()
	{
		if ((userName.getText()).length() > 9){
			JOptionPane.showMessageDialog(null, "User name too long! 8 characters limit.", 
				 					"Error!", JOptionPane.ERROR_MESSAGE);
			return false;
		} 
		
		if ((userName.getText()).length() < 8){
			JOptionPane.showMessageDialog(null, "User name too short! Must be 8 characters.", 
				 					"Error!", JOptionPane.ERROR_MESSAGE);
			return false;
		}
		
		uId = userName.getText();
		return true;
	}
	/**
	* returns true if the password field has exactly 8 chracters in length
	* returns false otherwise.
	*/
	public boolean preCheckPssWd()
	{
		if ((passWord.getPassword()).length > 8){
			JOptionPane.showMessageDialog(null, "Password too long!", 
				 					"Error!", JOptionPane.ERROR_MESSAGE);
			return false;
		}

		if ((passWord.getPassword()).length < 8){
			JOptionPane.showMessageDialog(null, "Password too short!", 
				 					"Error!", JOptionPane.ERROR_MESSAGE);
			return false;
		}
		
		psswd = new String(passWord.getPassword());
		return true;
	}
}
