package cs297.xmldb.src.cm;
import cs297.xmldb.src.sm.*;
import cs297.xmldb.src.bm.*;
import cs297.xmldb.src.server.*;
import java.io.*;
import java.util.*;

/**
 @Date: Spring & fall 2004
 @author: Thien An Nguyen, 
 A logical representation for a group node.
 */
public class ThGroup extends AThNode implements Serializable
{
  public ThAttributeNode attribute; 
  public ArrayList children;
  public String txId;
  int pageNum;
  
  public ThGroup()
  {}
  
  public ThGroup(int id, String attr, String dat, int pnum)
  {
     super(id);
 
     attribute = new ThAttributeNode(attr,dat);
     children = new ArrayList();
     pageNum = pnum;
  }     
  
  public ThGroup(int id, String dummy1, String dummy2)
  {
     super(id);
 
     attribute = new ThAttributeNode(dummy1,dummy2);
     children = new ArrayList();
     pageNum = -1;
  }     
    
  public ThGroup(int id, byte[] arr, int offset)
  {
     super(id);
     final int BYTE4 = 4;
     int len = 0;
     String data = "";
     int childcount = 0;
    
     int pos = offset;   /* get local copy of offset */
     
     /* decode page number */
     len = ThUtilities.hexToNumber(arr, pos);
     pos += BYTE4; //len of tag
     data = ThUtilities.hexToData(arr, pos, len);
     pos += len;
     pageNum = Integer.parseInt(data);
    
     /* decode attribute node */
     attribute = new ThAttributeNode(arr, pos);
     pos += attribute.size();
     
     /* decode number of children */
     len = ThUtilities.hexToNumber(arr, pos);
     pos += BYTE4; //len of tag
     data = ThUtilities.hexToData(arr, pos, len);
     pos += len;
     childcount = Integer.parseInt(data);
     
     /* array of node id */
     children = new ArrayList();
     for (int i = 0; i < childcount; i++)
     {
        len = ThUtilities.hexToNumber(arr, pos);
        pos += BYTE4;
        data = ThUtilities.hexToData(arr, pos, len);
        pos += len;
        children.add(data);
     }
  }  
  
 /**
  convert in memmory structure of this node
  to a sequence of byte array on a disk block/chunck 
  @Receive: the disk block, and offset for this node.
  @Returns: Return Code
  */
  public int toHex(char[] arr, int offset)
  {
     int pos = offset;
     IThNode child;
     int len = 0;
     String data = "";
     int id;
     
     /* Group or Leave node. 1 for G. 0 otherwise */
     BitSet byt = new BitSet();
     byt.set(6);  /* 7 - Proxy, 6 - Group */
     pos += ThUtilities.BitsetToChar(arr, byt, pos); 
     
     /* page num */
     data = Integer.toString(pageNum);
     pos += ThUtilities.numberToHex(arr, data.length()*2, pos);
     pos += ThUtilities.dataToHex(arr, data, pos);
     
     /* atribute data */
     pos += attribute.toHex(arr, pos);
     
     /* number of child */
     int numchild = children.size();
     data = Integer.toString(numchild);
     pos += ThUtilities.numberToHex(arr, data.length()*2, pos);
     pos += ThUtilities.dataToHex(arr, data, pos);
     
     /* id of childrend */
     for (int i = 0; i < numchild; i++)
     {
        child = (IThNode) children.get(i);
        id = child.getId();
        pos += ThUtilities.numberToHex(arr, data.length()*2, pos);
        pos += ThUtilities.dataToHex(arr, data, pos);
     }
     
     return pos - offset;
  }
  
  public int toHex(char[] arr, int offset, int idoffset)  
  {
     return toHex(arr, offset);
  }  
  
  public IThNode getChild(int idx) 
  {
     return (IThNode) children.get(idx); //ThGroup
  }
  
  public ThAttributeNode getAttributeNode()
  {
     return attribute;
  }
  
  public ArrayList getChildren()
  {
     return children;
  }
  
  public int getChildrenCount() 
  {
     return children.size();
  }
  
  public void insert(IThNode node) 
  { }
  
  public int getId()
  {
     return super.getId();
  }
  
 /**
  Returns the representation of this node as a string.
  @Receive: ThSystem where global variables resign
  @Returns: The string representation for this node
  */
  public String toString(ThSystem sys)
  {
     String value="";
    
     if (!isDeleted())
     {
        IThPage page = sys.getBM().getPage(Integer.toString(pageNum), sys);
        ArrayList data = page.getData(iD);
      
        value = (String) data.get(0);
        value += "="+(String) data.get(1);
     }
    
     return value;
  }
   
  public boolean hasData()
  {
     if (attribute != null)
        return true;
   
     return false;
  }
 
 /**
  Check to see if this node has children or not
  @Returns: True if there is at least 1 child. False otherwise.
  */
  public boolean hasChildren()
  {
     if (children.size() > 0)
        return true;
   
     return false;
  }
 
 /**
  size(id) + size(attr) + size(page num) + size(num child)
  + offset pointer with in the page + size (childid)
  */
  public int size() 
  {
     int pnsize = 4 + (Integer.toString(pageNum)).length()*2;
     int childcount = children.size();
     int childnumsize = 4 + (Integer.toString(childcount)).length()*2;
     int childsize = 0;

     int id;
     for (int i = 0; i < childcount; i++)
     {
        id = ((IThNode)children.get(i)).getId();
        childsize += 4 + (Integer.toString(id)).length()*2;
     }
     
     return (super.size() + pnsize + attribute.size()
                + childnumsize + childsize); 
  }
 
  public int getPageNum()
  {
     return pageNum;
  }
 
  public boolean isMatch(String key)
  {
     return super.isMatch(key);
  }
 
   public void addChild(AThNode node)
   {
      children.add(node);
   }
}