package cs297.xmldb.src.cm;
import cs297.xmldb.src.sm.*;
import cs297.xmldb.src.bm.*;
import cs297.xmldb.src.server.*;
import java.io.*;
import java.util.*;

/**
 @Date: Spring & fall 2004
 @author: Thien An Nguyen, 
 A logical representation for a leave node.
 */
public class ThLeave extends AThNode implements Serializable
{
 int pageNum;
 String txId;
 
 public ThLeave()
 {}
 
 public ThLeave(int id, int pnum, String txid, ThSystem sys)
 {
    super(id);   
    pageNum = pnum;
    txId = txid;
 }
 
 public ThLeave(int id, int pnum, ThSystem sys)
 {
    super(id);   
    pageNum = pnum;
 }
 
 /**
  This constructor create a leave object from a byte array 
  of HEX characters.
  @param: id of the leave,
          the byte array,
          position where to start decode
  */  
  
  public ThLeave(int id, byte[] arr, int offset)
  {
     super(id);
     
     final int BYTE4 = 4;
     int len = 0;
     String data = "";
     int childcount = 0;
    
     int pos = offset;   /* get local copy of offset */
     
     /* decode page number */
     len = ThUtilities.hexToNumber(arr, pos);
     pos += BYTE4; //len of tag
     data = ThUtilities.hexToData(arr, pos, len);
     pos += len;
     pageNum = Integer.parseInt(data);
  }  
  
 /**
  convert in memmory structure of this node
  to a sequence of byte array on a disk block/chunck 
  @Receive: the disk block, and offset for this node.
  @Returns: Return Code
  */
  public int toHex(char[] arr, int offset)
  {
     int pos = offset;
     int len = 0;
     String data = "";
     
     /* Group or Leave node. 1 for G. 0 otherwise */
     BitSet byt = new BitSet(); /* 7 - Proxy, 6 - Group, other wise - Leave*/
     pos += ThUtilities.BitsetToChar(arr, byt, pos); 
     
     /* page num */
     data = Integer.toString(pageNum);
     pos += ThUtilities.numberToHex(arr, data.length()*2, pos);
     pos += ThUtilities.dataToHex(arr, data, pos);
     
     return pos - offset;
  }  
 
  public int toHex(char[] arr, int offset, int idoffset)
  {
     return toHex(arr, offset);
  } 
  
  public int getId()
  {
     return super.getId();
  }
 
  public int getPageNum()
  {
     return pageNum;
  }
 
  public String getTxId()
  {
     return txId;
  }
 
  public IThNode getChild(int idx) 
  {
     return null;   // no more child 
  }
 
  public ArrayList getChildren()
  {
     return null; 
  }
  
  public int getChildrenCount()
  {
     return 0;
  } 
 
  public void insert(IThNode node)
  {
  }
  
 /**
  Returns the string that represents for the data in this leave.
  @Receive: ThSystem where global variables resign.
  @Returns: The string
  */ 
  public String toString(ThSystem sys)
  {
     ThBufferMgr BM = sys.getBM();
     IThPage page = sys.getBM().getPage(Integer.toString(pageNum), sys);
     ArrayList data = page.getData(iD);
      
     return (String) data.get(0);
  }

 public boolean hasData()
 {
    return true;
 }
 
 public int size()
 { //2bytes offset + size data (pnum)
    return super.size() + 4 + (Integer.toString(pageNum)).length()*2;
 }
 
 public boolean isMatch(String key)
 {
    return super.isMatch(key);
 }
}
