package cs297.xmldb.src.cm;
import java.io.*;
import java.util.*;
import java.lang.*;
import cs297.xmldb.src.sm.*;
import cs297.xmldb.src.server.*;

/**
 @Date: Spring & fall 2004
 @author: Thien An Nguyen, 
 This class is used to parsing through input file 
 for our DDL/DML syntax errors.
 */
public class ThParser 
{      
 /**
  This fuction error check for grammar and syntax error for 
  statements, commands in the input file.
  @receive: Properties obj to check for Error codes
            defined in the properties file.
  @return:  An object called ThPlan which is the result 
            of running through this error check function.
  */
  public static ThPlan errorCheckDML(Properties prop, ThSystem sys)
  {  
     sys.out("Checking DDL/DML error(s) ...");
   
     String testcase = (String)prop.get("testFile");
     //break DML plan into stmts and/or commands
     ArrayList lst = scanForStatements(testcase, sys);    
     if (lst == null)
        System.exit(1);
   
     //Now, check for syntax & grammar
     ArrayList stmts = new ArrayList(); 
     boolean succ = false;   
     ArrayList databases = new ArrayList();
     ArrayList xmldocs = new ArrayList();
     ArrayList params;
     String word = "",
          txid;
     ThStatement stmt = null;
   
     //go through each word in each stmt
     for (int i=0; i<lst.size(); i++)
     {
        params = new ArrayList();
        String temp = ((StringBuffer)lst.get(i)).toString();
        StringTokenizer line = new StringTokenizer(temp);
        //parse through each statement from the lst
        txid = line.nextToken(); 
     
        while (line.hasMoreTokens())
        {    
           word = (line.nextToken()).toUpperCase();
           if (word.indexOf("CRASH") >= 0)  
           {
              stmt = new ThStatement(txid,"CRASH", params, 
                     (String)prop.get("EC0"));  
           }
           else if(word.equals("SELECT"))    //SELECT    
              stmt = parseSelect(txid, line, xmldocs, prop);       
           else if (word.equals("INSERT"))   //INSERT
              stmt = parseInsert(txid, line, databases, xmldocs, prop);
           else if (word.equals("UPDATE"))   //UPDATE
              stmt = parseUpdate(txid, line, xmldocs, prop);
           else if (word.equals("DELETE"))   //DELETE      
              stmt = parseDelete(txid, line, xmldocs, prop);
           else if (word.equals("CREATE"))   //CREATE      
              stmt = parseCreate(txid, line, databases, xmldocs, prop);                                             
           else if (word.equals("LOAD"))    
              stmt = parseLoad(txid, line, databases, xmldocs, prop);    
           else if (word.equals("COMMIT"))    //COMMIT
              stmt = new ThStatement(txid,"COMMIT", params, (String)prop.get("EC0"));
           else if (word.equals("ROLLBACK"))  //ROLLBACK
              stmt = new ThStatement(txid,"ROLLBACK", params, (String)prop.get("EC0"));
           else if (word.equals("DROP"))        
              stmt = parseDrop(txid, line, databases, xmldocs, prop);
           else if(word.indexOf("PRINTTREE") >= 0)
           {
              params.add(temp.substring(temp.indexOf('(')+1,
                                   temp.indexOf(')')));
              stmt = new ThStatement(txid, "PRINTTREE", params, (String)prop.get("EC0"));         
           }
           else if(word.indexOf("ARCHIVE") >= 0)
              stmt = parseArchive(txid, line, databases, prop);
           else if(word.equals("DISPLAY"))      //DISPLAY 
              stmt = new ThStatement(txid,"DISPLAY", params, (String)prop.get("EC0"));         
           else
           {            // Syntax Error
              ThStatement obj = new ThStatement(txid, "UNKNOWN COMMAND OR SQL", 
                                   params, (String)prop.get("UNKNOWN_COMMAND"));
           }
       }//end while  
       stmts.add(stmt); 
     }//end for
     return new ThPlan(lst, stmts, (String)prop.get("EC0"));
 }  
  
 /** 
  It parses the input file for statements.
  A statement is terminated by ';'
  Thus, everytime it sees a ';', it assumed 
  the previous text up til then is a statement.
  @params: name of the input flat file where DDL/DML and 
           commands are contained.
  @returns: ArrayList of statements for next process
  */      
 private static ArrayList scanForStatements(String fileName, ThSystem sys)
 {
    ArrayList cmdList = new ArrayList();
    try
    {
       File file = new File(fileName);
       if (!file.exists())
       {
          sys.out("File name Error!\nExited!");
          return null;
       }
    
       DataInputStream in = new DataInputStream(
                           new BufferedInputStream(
                               new FileInputStream(fileName)));
       /*read in line by line*/
       String inLine = new String();
       StringBuffer cmd = new StringBuffer();
    
       while((inLine = in.readLine())!=null)
       {
          if (inLine.indexOf("/*") >= 0) /* this is a comment do nothing */
             ;
          else if (inLine.indexOf(';') < 0)
             cmd.append(inLine);    
          else                   /*one or more ';'           */
          {                      /* found on the same line   */
             int idx = inLine.indexOf(';');
             while (idx > 0)
             {                   /*ignore if ';' is at the beginning */
                cmd.append(inLine.substring(0,idx));  /* of the stmt */
                cmdList.add(cmd);           /* append the first stmt */
                cmd = new StringBuffer();
          
                inLine = new String(inLine.substring(idx+1, inLine.length()));
                idx = inLine.indexOf(';');
             }
          }   
       } //end while
       in.close();    
    }
    catch(Exception e)
    {
       e.printStackTrace();
    }
    return cmdList;
 }
 
 /**
  There should be exactly 3 tokens for a select statement.
  If the token count fail => statement error
  For example: SELECT "/G[*]" FROM XML1DOC;
  The above statment has 4 tokens but SELECT 
  already PARSED by scanForStatement
  @receive: tokenized statement, 
             existing xmldocs, 
             and properties list for error checking purpose
  @return: Insert statement
  */
 private static ThStatement parseSelect(String txid, StringTokenizer line, 
                          ArrayList xmldocs, Properties prop)
 {
    ArrayList params = new ArrayList();
    String word;
   
    /* There should be 3 tokens left for a select statement */
    if (line.countTokens() != 3)
       return new ThStatement(txid, "Token Error!", params, 
                              (String)prop.get("TOKEN_cOUNT_ERROR"));
     
    /* parse 2nd token for predicate */
    word = line.nextToken();
    ArrayList paths = new ArrayList();
    paths.add(word.substring(word.indexOf('"')+1, 
                            word.lastIndexOf('"')));
    params.add(paths);
   
    /* Check if the 3rd token is FROM */
    word = line.nextToken();
    if (!word.equals("FROM"))
       return new ThStatement(txid, "In-complete SELECT Statement!", 
                            params,(String)prop.get("EC3"));
       
    /* 4th token should be xml doc name */
    word = line.nextToken(); 
    if (!xmldocs.contains(word))
       return new ThStatement(txid,"Non existing database or document: "+word, 
                            params,(String)prop.get("EC3"));
    ArrayList docs = new ArrayList();
    docs.add(word);
    params.add(docs);
    
    return new ThStatement(txid, "SELECT", params, (String)prop.get("EC0"));
 }
 
 /**
  There should be at least 4 tokens for a select statement.
  If the token count fail => statement error
  receive: transaction id, StringTokenizer to be continue to 
           parse, list of databases name currently created in
           the system, xmldocnames currently created in the 
           system, and properties obj
  @return: The insert statement
  */
 private static ThStatement parseInsert(String txid, StringTokenizer line,
                  ArrayList databases, ArrayList xmldocs, Properties prop)
 {
    ArrayList params = new ArrayList();
    String word;
    ThStatement stmt = null;   
    //There should be at least 4 tokens left for a insert statement
    if (line.countTokens() < 4)
       return new ThStatement(txid,"Token Error!", params, 
                              (String)prop.get("TOKEN_cOUNT_ERROR"));   
    //1st tokend, keyword INTO
    word = line.nextToken();
    if(!word.equals("INTO"))
       return new ThStatement(txid,"Error! Missing INTO Keyword.", 
                            params,(String)prop.get("EC2"));   
    //2nd token, parse for docname
    word = line.nextToken();
    if(!xmldocs.contains(word))
       return new ThStatement(txid,"XMLDOC "+word+" NOT FOUND!", 
                            params,(String)prop.get("EC2"));    
    params.add(word);       
    //3rd token, should be PATH
    word = line.nextToken();
    if(word.indexOf("PATH")<0)
       return new ThStatement(txid,"Error! Missing PATH Keyword.", 
                            params,(String)prop.get("EC2"));   
    params.add(word.substring(word.indexOf('"')+1, 
                             word.lastIndexOf('"')));
    //4th token, should be VALUES
    word = line.nextToken();
    if (word.indexOf("VALUES")<0)
       return new ThStatement(txid,"Error! Missing VALUES Keyword.", 
                            params,(String)prop.get("EC2")); 
   
    ArrayList values = new ArrayList(); 
    while(line.hasMoreTokens())
       word += (" "+line.nextToken());
        
    if(word.indexOf("<XML") >= 0 || word.indexOf("<G") >= 0)
    {
       word = word.substring(word.indexOf("<"),
                            word.lastIndexOf(">"));
       values.add(word);
       params.add(values);
    }
    else if(word.indexOf("\", ")<0)
    {   //1 value added into specified path 
       word = word.substring(word.indexOf("<L")+3,
                            word.lastIndexOf("</L"));
       values.add(word);
       params.add(values);
    }
    else
    {   //more than 1 values added to the specified path 
       while (word.indexOf("<L") >= 0)
       {
          int idx = word.indexOf("</L");
          String tmp = word.substring(word.indexOf("<L")+3,idx);
          word = word.substring(idx+4);
          values.add(tmp);  
       } //end while
       params.add(values);                           
    } 
    return stmt = new ThStatement(txid,"INSERT", params, (String)prop.get("EC0")); 
 }
 
 /**
  There should be at least 5 tokens for update statement.
  If the token count fail => statement error
  @receive: transaction id, input data to be parsed,
            list of xml documents currently in the system
  @return: Update statement
  */
  private static ThStatement parseUpdate(String txid, StringTokenizer line,
                     ArrayList xmldocs, Properties prop)
  {  
     ArrayList params = new ArrayList();
     String word;
     ThStatement stmt = null;
   
     if (line.countTokens() < 5)
        return new ThStatement(txid,"Token Error!", params, 
                             (String)prop.get("EC2"));
  
     //1st token, xml doc-name
     word = line.nextToken(); 
     if(!xmldocs.contains(word)) //make sure docname already exist
        return new ThStatement(txid,"Document name is not exist!", 
                     params,(String)prop.get("TOKEN_cOUNT_ERROR"));   
   
     params.add(word);
     //2nd token, keyword SET        
     word = line.nextToken(); 
     if (!word.equals("SET"))
        return new ThStatement(txid,"Syntax Error!", params,
                              (String)prop.get("EC2"));   
   
    //3rd token, VALUE clause
    word = line.nextToken(); 
    if (word.indexOf("VALUE")<0)
       return new ThStatement(txid,"Syntax Error!", params, 
                            (String)prop.get("EC2"));
     
    while(word.lastIndexOf("</L>")<0)
       word += (" "+line.nextToken());
                          
    word = word.substring(word.indexOf("<L")+3, word.lastIndexOf("</L"));
    params.add(word);  

    //4th token, where clause
    word = line.nextToken();
    if (word.indexOf("WHERE") < 0 )
       return new ThStatement(txid,"Syntax Error!", params, 
                                (String)prop.get("EC2")); 
     
    word = line.nextToken();     //5th token
    if (word.indexOf("PATH") <0 ) 
       stmt = new ThStatement(txid,"Update Missing path(s)!",
                            params,(String)prop.get("EC2"));
   
    ArrayList values = new ArrayList();
    while (true)
    {
       word = word.substring(word.indexOf('"')+1,
                           word.lastIndexOf('"'));
       values.add(word);
       params.add(values);
       if (line.hasMoreTokens())
          word = line.nextToken();
       else
          break;                        
    }
   
    return stmt = new ThStatement(txid,"UPDATE", params, (String)prop.get("EC0"));   
  }
  
 /**
  There should be at least 4 tokens for DELETE statement.
  If the token count fail => statement error
  @receive: transaction id, input data to be parsed,
             list of xml documents currently in the system
  @return: Delete statement
  */
  private static ThStatement parseDelete(String txid, StringTokenizer line,
                                ArrayList xmldocs, Properties prop)
  {
     ArrayList params = new ArrayList();
     String word;
   
     if (line.countTokens() < 4)
        return new ThStatement(txid,"Token Error!", params,
                             (String)prop.get("TOKEN_cOUNT_ERROR"));
   
    //1st token
    word = line.nextToken();
    if(!word.equals("FROM"))
       return new ThStatement(txid,"Error! Missing FROM Keyword.", 
                              params,(String)prop.get("EC2")); 
   
    //2nd token
    word = line.nextToken();
    if(!xmldocs.contains(word)) //make sure docname already exist
       return new ThStatement(txid,"Document name is not exist!", 
                              params,(String)prop.get("EC2")); 
   
    params.add(word);   
    //3rd token
    word = line.nextToken();
    if (!word.equals("WHERE"))
       return new ThStatement(txid,"Error! Missing WHERE Keyword.", 
                              params,(String)prop.get("EC2")); 
   
    //4th token
    word = line.nextToken();
    if (word.indexOf("PATH")<0)
       return new ThStatement(txid,"Missing path(s)!", params,
                              (String)prop.get("EC2"));
                 
    //now, parse path(s)
    ArrayList values = new ArrayList();
    while (true)
    {
       word = word.substring(word.indexOf('"')+1,
                             word.lastIndexOf('"'));
       values.add(word);
       params.add(values);
       if (line.hasMoreTokens())
           word = line.nextToken();
       else
           break;                         
    }
                   
    return  new ThStatement(txid,"DELETE", params, (String)prop.get("EC0"));
  }
  
 /**
  There should be at least 2 tokens for CREATE statement.
  If the token count fail => statement error
  @receive: transaction id, input data to be parsed,
             list of xml documents currently in the system
  @return: Create statement
  */
  private static ThStatement parseCreate(String txid, StringTokenizer line, 
                   ArrayList databases, ArrayList xmldocs, Properties prop)
  {
     ArrayList params = new ArrayList();
     String word;
     ThStatement stmt = null;
   
     if (line.countTokens() < 2)
        return new ThStatement(txid,"Token Error!", params,
                            (String)prop.get("TOKEN_cOUNT_ERROR"));
   
     //1st token, name of xmldoc or database
     word = line.nextToken();    
     if(word.equals("XMLDOC"))
     {   //xmldoc
        word = line.nextToken(); //2nd token, doc name
        params.add(word);
        xmldocs.add(word);
        if(!line.hasMoreTokens())
           return new ThStatement(txid,"In-complete Statement!", 
                              params, (String)prop.get("EC2"));
             
        word = line.nextToken(); 
        if(!word.equals("IN"))
           return new ThStatement(txid,"Syntax Error!", params,
                              (String)prop.get("EC2"));
       
        if(!line.hasMoreTokens())
           return new ThStatement(txid,"Missing Database Name!", params,
                              (String)prop.get("EC2"));
           
        word = line.nextToken();
        if (databases.contains(word))
        {
           params.add(word);
           stmt = new ThStatement(txid,"CREATE XMLDOC", params, 
                              (String)prop.get("EC0"));
        }
        else
           return new ThStatement(txid,"DATABASE "+word+" NOT FOUND!", 
                              params, (String)prop.get("EC2"));     
     }
     else if(word.equals("DATABASE"))
     { //if database
        word = line.nextToken(); //database name
        if (!databases.contains(word))
        {
           params.add(word);
           databases.add(word);
           stmt = new ThStatement(txid,"CREATE DATABASE", params, 
                              (String)prop.get("EC0"));
        }
        else
           return new ThStatement(txid,"Database Name already exist!", 
                              params, (String)prop.get("EC2"));
     }
     else  //un-expected
        return new ThStatement(txid,"CREATE DATABASE/XMLDOC?", params,
                            (String)prop.get("EC2"));
     
     return stmt;
  }   

 /**
  There should be at least 3 tokens for LOAD statement.
  If the token count fail => statement error
  @receive: transaction id, input data to be parsed,
            list of database's names currently in the system
            list of xml documents currently in the system
  @return: Load statement
  */
  private static ThStatement parseLoad(String txid, StringTokenizer line, 
             ArrayList databases, ArrayList xmldocs, Properties prop)
  {
     ArrayList params = new ArrayList();
     String word;
     ThStatement stmt = null;
   
     if (line.countTokens() != 3)
        return new ThStatement(txid,"Token Error!", params,
                            (String)prop.get("TOKEN_cOUNT_ERROR"));
     //1st token
     word = line.nextToken();      // filename followed
     word = word.substring(word.indexOf('"')+1, 
                         word.lastIndexOf('"'));
     params.add(word);   
     //2nd token
     word = line.nextToken();
     if(!word.equals("INTO"))
        return new ThStatement(txid,"Syntax Error!", params,
                            (String)prop.get("EC2"));   
     //3rd token, doc name
     word = line.nextToken();
     if(!xmldocs.contains(word))
        return new ThStatement(txid,"XmLDOC "+word+" NOT FOUND!", 
                            params, (String)prop.get("EC2")); 
     params.add(word);
   
     return stmt = new ThStatement(txid,"LOAD", params, (String)prop.get("EC0")); 
  }
  
 /**
  There should be exact 2 tokens for LOAD statement.
  If the token count fail => statement error
  This function won't check for database or 
  xmsdoc existence because one could drop a 
  non-existing database or xmldoc before create one.
  
  @receive:transaction id, input data to be parsed,
            list of xml documents currently in the system
  @return: Drop statement
  */
  private static ThStatement parseDrop(String txid, StringTokenizer line, 
                ArrayList databases, ArrayList xmldocs, Properties prop)
  {
     ArrayList params = new ArrayList();
     String word;
     ThStatement stmt = null; 
    
     if (line.countTokens() != 2)
        return new ThStatement(txid,"Token Error!", params,
                             (String)prop.get("TOKEN_cOUNT_ERROR"));
   
     //1st token
     word = line.nextToken();
     if(word.equals("XMLDOC"))
     {
        word = line.nextToken();   //doc name
        params.add(word);
        xmldocs.remove(word);
        stmt = new ThStatement(txid,"DROP XMLDOC", params, (String)prop.get("EC0"));
     }
     else if (word.equals("DATABASE"))
     {
        word = line.nextToken();   //database name
        params.add(word);
        databases.remove(word);
        stmt = new ThStatement(txid,"DROP DATABASE", params, (String)prop.get("EC0"));
     }
     else
        return new ThStatement(txid,"Unknown Command!", params,(String)prop.get("EC2")); 
   
     return stmt;
  }
  
 /**
  Parsing handler for the Archive statement
  @receive:transaction id, input data to be parsed,
             list of xml documents currently in the system
  @return: Drop statement
  */  
  private static ThStatement parseArchive(String txid, StringTokenizer line, 
                                 ArrayList databases, Properties prop)
  {
     ArrayList params = new ArrayList();
     
     if (line.countTokens() != 1)
        return new ThStatement(txid,"Token Error!", params,
                             (String)prop.get("TOKEN_COUNT_ERROR"));
    
     String word = line.nextToken();
     int idx = databases.indexOf(word);
    
     if (idx < 0)
        return new ThStatement(txid,"Database Not Found!", params,
                             (String)prop.get("DB_NOT_FOUND"));
    
     params.add(word);
     return new ThStatement(txid,"ARCHIVE", params,
                             (String)prop.get("EC0"));
  }
}
