package cs297.xmldb.src.cm;

import java.io.*;
import java.util.*;
import java.lang.*;
import cs297.xmldb.src.cm.*;
import cs297.xmldb.src.bm.*;
import cs297.xmldb.src.sm.*; 
import cs297.xmldb.src.server.*;  

/**
 This class handles lots of important jobs for storage management layer 
 such as convert data from one to another form.
 It also handles Image copy and retreive image copy.
 @Cautions: Utilities ahead!
 @Date: Spring 2004
 @author: Thien An Nguyen, 
 */
public class ThUtilities
{  
   /**
   Convert a BitSet (8 bits) into half byte in hex.
   @Receive: The array of character and the offset for the 
             converted char to be on,
             The BitSet to be converted.
   @Return: the length of converted data, which is only half byte.
   */
  public static int BitsetToChar(char[] arr, BitSet set, int offset)
  {
     int result = 0;
    
     if (set.get(1))
        result += Math.pow(2,6);
     if (set.get(2))
        result += Math.pow(2,5);
     if (set.get(3))
        result += Math.pow(2,4);
     if (set.get(4))
        result += Math.pow(2,3);
     if (set.get(5))
        result += Math.pow(2,2);
     if (set.get(6))
        result += Math.pow(2,1);
     if (set.get(7))
        result += 1;
    
     arr[offset] = toHexChar(result % 16);    
     return 1;
  }
  
  /**
   Convert half byte character to a BitSet
   @Receive: The byte array that contains the character,
             the offset of the character.
   @Return: Converted BitSet
   */  
  public static BitSet CharToBitset(byte[] bytes, int offset)
  {
     int result = (char) bytes[offset];
     BitSet s = new BitSet(8);
    
     if (result == '1')
        s.set(7);
     else if (result == '2')
        s.set(6);
     else if (result == '3')
     {
        s.set(7);
        s.set(6);
     }
     else if (result == '4')
        s.set(5);
     else if (result == '5')
     {
        s.set(1);
        s.set(5);
     }        
    
     return s;
  }  
  
  /**
   Convert a half byte to hex
   @Receive: the ascii
   @Return:  the hex character 
   */  
  public static char toHexChar(int b)
  {
     char[] conversion = {'0','1','2','3','4','5','6','7','8','9'};
    
     if (b < 10) 
        return conversion[b];
     else if(b == 10)
        return 'a';
     else if(b == 11)
        return 'b';
     else if(b == 12)
        return 'c';
     else if(b == 13)
        return 'd';
     else if(b == 14)
        return 'e';
     else 
        return 'f';
  }
  
  /**
   Convert a half byte hex into a number
   @Receive: the half byte number
   @Return:  the integer from 0 - 15.
   */  
  public static int toAsciiInt(int num)
  {
     char b = (char) num;
     if (b == 'a')
        return 10;
     else if(b == 'b')
        return 11;
     else if(b == 'c')
        return 12;
     else if(b == 'd')
        return 13;
     else if(b == 'e')
        return 14;
     else if(b == 'f')
        return 15;
     else 
        return Character.digit(b,10);
  }
  
  /**
   Convert 1 byte of hex into character data
   @Receive: byte array and the offset of the first half
   @Return:  a single character data.
   */  
  public static char byteHexToData(byte[] bytes, int offset)
  {
     char data;
     int num;
     int c1 = toAsciiInt(bytes[offset]);
     int c2 = toAsciiInt(bytes[offset+1]);
     
     num = (byte) c1 << 4;
     num += (byte) c2;
     data = (char) num;
     
     return data;
  }  
  
  /**
   Convert one character ascii into 1 byte hex.
   @Receive: the hex character array and the position to put the 
            first half of the converted byte.
   @Return:  the length of converted data, which is one byte.
   */  
  public static int byteDataToHex(char[] arr, byte ascii, int pos)
  {
     int a1 = ascii >> 4;
     int a2 = ascii % 16;
     
     arr[pos] = toHexChar(a1);
     arr[pos+1] = toHexChar(a2);

     return 2;
  }  
    
  /**
   This function goes through each character in
   the passed string data and convert them into hex.
   Then put each byte of the converted character on the character array.
   @Receive: character array, the string to be convert
             and the offset
   @Return:  the length of the converted data
   */  
  public static int dataToHex(char[] arr, String data, int offset)
  {
     int pos = offset;
     byte[] bytes = data.getBytes();
     int len = bytes.length;
    
     for (int i=0; i<len; i++)
        pos += byteDataToHex(arr, bytes[i], pos);
    
     return pos-offset;
  }
  
  /**
   Convert an integer to hex (2 bytes).
   @Receive: the character arry and the number to be conveted,
             and the position to put the result in arr.
   @Return:  length of converted data.
   */  
  public static int numberToHex(char[] arr, int number, int pos)
  {
     if (number < 0)
     {
        arr[0] = arr[1] = arr[2] = arr[3] = 'F';
        return 4;
     }
     
     int a1 = number, a2 = 0;
     int idx = pos+3;
     
     while ((a1 >> 4) >= 1)
     {
        a2 = a1 % 16;
        a1 = a1 >> 4;
        arr[idx--] = toHexChar(a2);
     }
     arr[idx--] = toHexChar(a1);
     
     //fill the least significance with 0s
     int k=pos;
     while (k <= idx)
     {
        arr[k++] = '0';
     }

     return 4;
  }
  
  /**
   Convert 1 byte of hex back into character data
   @Receive: byte array and the offset of the starting point,
             and the length of the data need to be converted.
   */  
  public static String hexToData(byte[] bytes, int offset, int length)
  {
     String data = "";
     int pos = offset;
    
     for (int i=0; i<length; i += 2)
     {
        data += byteHexToData(bytes, pos);
        pos += 2;
     }
    
     return data;
  }  
 
  /**
   Convert 2 bytes of hex back into a number.
   @Receive: byte array and the offset
   @Return:  the number.
   */  
  public static int hexToNumber(byte[] bytes, int offset)
  {
     double num;
     
     int c0 = toAsciiInt(bytes[offset]);
     int c1 = toAsciiInt(bytes[offset+1]);
     int c2 = toAsciiInt(bytes[offset+2]);
     int c3 = toAsciiInt(bytes[offset+3]);
     
     num = c0*Math.pow(16,3);
     num += c1*Math.pow(16,2);
     num += c2*16;
     num += c3;
     
     return (int) num;
  }
  
  /**
   Make an image copy of a database 
   @Receive: the database to be making image of
             ThSystem where a global time clock
             is ticking every time a request for time
   @Return:  Return code.
   */  
  public static int imageCopy(ThDB db, ThSystem sys)
  {
     String dbname = db.getName();
     Properties prop = sys.getProperties();
     String dir = (String) prop.get("archiveDirectory");
     String fname = dir+dbname+".dat"+sys.getCurrentTimeStamp();
     try
     {
        ObjectOutputStream out = new ObjectOutputStream(
                                new FileOutputStream(fname));  
      
        out.writeObject(db);       
        /* call system to discard log records up to here */
        out.close();
        return 0;
     }
     catch(Exception e)
     {
        e.printStackTrace();
     }      
    
     return -1;
  }
  
  /**
   Retreive the last image copy of the data's name.
   @Receive: name of the database and ThSystem to get time
   @Return: the previous state of the database.
   */  
  public static ThDB getImageCopy(String dbname, ThSystem sys)
  {
     Properties prop = sys.getProperties();
     String dir = (String) prop.get("archiveDirectory");
     String fname = dir+dbname+".dat"+sys.getLastTimeStamp();
     try
     {
        ObjectInputStream inStream = new ObjectInputStream(
                                  new BufferedInputStream(
                                  new FileInputStream(fname)));      

       return (ThDB) inStream.readObject();   
     }
     catch(Exception e)
     {
        e.printStackTrace();
     }      
    
     return null;
  }  
}