package cs297.xmldb.src.cm;

import java.io.*;
import java.util.*;
import cs297.xmldb.src.bm.*;
import cs297.xmldb.src.sm.*;
import cs297.xmldb.src.server.*;

/**
  COPYRIGHT (C) 2004 Thien An Nguyen. All Rights Reserved.
  Contains the logical structure and operations on xml document. 
  @Date: Spring & fall 2004
  @author: Thien An Nguyen, 
 */
public class ThXMLDoc implements Serializable
{
   String spaces = " "; //for print Tree alignment
   String namE = null;
   ThGroup root; //logical tree root
   int rootPageNum; //of the root page
   ArrayList strucPages;
   static final boolean LOG_YES = true;
   static final boolean LOG_NO = false;
   Stack pageSet = new Stack();
  
   /**
   Construct an empty xml document structure.
   @Receive: The name and ThSystem.
   @Returns: none (A ThXMLDoc is created).
   */
  public ThXMLDoc(String xmlname, ThSystem sys)
  {
     namE = new String(xmlname);
     int iD = ThSystem.getId();
     root = new ThGroup(iD, "ROOT", "ROOT");
     rootPageNum = (sys.getBM()).allocateStrucPageNum(sys);
     pageSet.push(Integer.toString(rootPageNum));
     strucPages = new ArrayList();
 
     if (!strucPages.contains(Integer.toString(rootPageNum)))
        strucPages.add(Integer.toString(rootPageNum));  
  }
  
  /**
  SELECT operation handler.
  @Return:  a list of objects selected
  @Receive: expression path
  */
  public Object select(String path, ThSystem sys)
  { 
     sys.out("Selecting from XMLDoc: "+namE);
     ArrayList resultSet = new ArrayList();
     //current position
     IThNode currPtr = (IThNode)((ThGroup)root).children.get(0);  
     int idx;
    
     while (true)
     {
        if (path.startsWith("/G"))
        {
           path = path.substring(2);        
        
          if (path.startsWith("["))
          { //indication selection of children
             if (path.indexOf("*")>=0)
             {  //all of the children
                resultSet = (ArrayList)((ThGroup)currPtr).getChildren();
             }
             else
             {              //indicate which child to be selected
                idx = Integer.parseInt(path.substring(path.indexOf("[")+1, 
                                                  path.indexOf("]")));
                resultSet.add(currPtr.getChild(idx-1));
                path = path.substring(path.indexOf("]")+1);
             }
          
             return resultSet;
          }
          else
          { //by default, first child in the group
             currPtr = (IThNode)currPtr.getChild(0);
          }         
        }
        else
        {
           sys.out("Unknown Path!");
           break;
        }
    }
    
    return resultSet;                                     
  }  
  
  /**
  * @Purpose: Handle DELETE operation
  * @Receive: Id of the record to be deleted
  *           The expression path.
  *           ThSystem sys
  * @Returns: Return code indicate of execution success for fail
  */
  public synchronized String delete(String txid, Object pat, 
                          ThSystem sys, boolean islog)
  {
     Properties prop = sys.getProperties();
     sys.out("Deleting from XMLDoc...");
    
     ArrayList arr = (ArrayList) pat;
     for (int k=0; k<arr.size();k++)
     {
        //current position
        IThNode currPtr = (IThNode)((ThGroup)root).children.get(0);  
        //ThGroup currPtr = ((ThGroup)root).children.get(0);
        Stack stack = new Stack();
        int idx = 0;
        String path = (String) arr.get(k);
        ThLogManager LM = sys.getLM();
        ThBufferMgr BM = sys.getBM();
        ThTransactionMgr TM = sys.getTM();
        String id;
        IThPage page;
        String lsn;
        String pnum;
        ArrayList children;
    
        while (true)
        {   //travese down the tree until the condition is matched
           if (path.startsWith("/G"))
           {
              path = path.substring(2);        
              stack.push(currPtr);
              if (path.startsWith("["))
              { //indication of which child
                 idx = Integer.parseInt(path.substring(path.indexOf("[")+1, 
                                 path.indexOf("]")));
                 idx -= 1; // array idx starts from 0
                 currPtr = (IThNode)currPtr.getChild(idx);
                 path = path.substring(path.indexOf("]")+1);
              }
              else //by default, first child in the group
                 currPtr = (IThNode)currPtr.getChild(0);
        
              if (path.equals(""))
              { //default, 1st child in the list of this group 
                 currPtr = (ThGroup)stack.pop();
                 children = ((ThGroup)currPtr).getChildren();  
                 id = Integer.toString(currPtr.getId());
                 pnum = Integer.toString(currPtr.getPageNum());
                 page = BM.getPage(pnum, sys); 
                 if (islog)
                    lsn = page.createDelLogRecord("UDELETE", txid, 
                                      Integer.parseInt(id), (String) arr.get(k),  
                                      namE, sys.getDB(namE), sys);    
                 ((IThNode)children.get(idx)).setDeleted(true);
                 break;        
              }      
          }
          else if (path.startsWith("/L"))
          {
             path = path.substring(2);        
             currPtr = (ThGroup)stack.pop();
        
             if (path.startsWith("["))
             { //indication of which child
                idx = Integer.parseInt(path.substring(path.indexOf("[")+1, 
                                 path.indexOf("]")));
                idx -= 1; // array idx starts from 0
          
                children = ((ThGroup)currPtr).getChildren();
                ThLeave leave = (ThLeave)children.get(idx);
                id = Integer.toString(leave.getId());
                pnum = Integer.toString(leave.getPageNum());    
                page = BM.getPage(pnum, sys);           
                lsn = page.createDelLogRecord("UDELETE", txid, leave.getId(), 
                            (String) arr.get(k),namE, sys.getDB(namE), sys);                   
                ((IThNode)children.get(idx)).setDeleted(true);
             }
             else
             { //by default, first child in the group
                children = ((ThGroup)currPtr).getChildren();
                ThLeave leave = (ThLeave)children.get(0);
                id = Integer.toString(leave.getId());
                pnum = Integer.toString(leave.getPageNum());    
                page = BM.getPage(pnum, sys);           
                lsn = page.createDelLogRecord("UDELETE", txid, Integer.parseInt(id), 
                            (String) arr.get(k),namE, sys.getDB(namE), sys);        
               ((IThNode)children.get(0)).setDeleted(true);
             }
             break;  
          }
          else
          {
             sys.out("something wrong!");
              break;
          }
       }
    }    
    return (String)prop.get("EC0");
  }
  
 /**
  * @Purpose: Handle INSERT operation
  * @Returns: Return Code 
  * @Receive: Transaction id,
  *           Path of the insert record,
  *           Leaves to be inserted,
  *           ThSystem
  */
  public String insert(String txid, String path, Object contents, 
                       ThSystem sys, boolean islog)
  { 
     sys.out("Inserting into XMLDoc...");
     Properties prop = sys.getProperties();
     //declaration & initialization
     String pat = path,
            line,
            attr,
            data;
     int idx,
         currpnum = rootPageNum;
     boolean insertgroup = false;
     ThGroup parent = (ThGroup) root;        
     ArrayList values = (ArrayList)contents;
     line = (String) values.get(0);

     //determind what kind of insert is this (G or L)
     if (line.indexOf("XML") >= 0 || path.equals(".") || path.indexOf("/L") < 0)
        insertgroup = true;
    
     //traverse the tree base on the path    
     if (path.equals("."))
        path = "";
     else
     {     
        parent = (ThGroup)parent.getChildren().get(0); //xml node
        while (!path.equals(""))
        {
           if (path.equals("/G") || path.equals("/L"))
              path = path.substring(2); 
           else if (path.startsWith("/G"))
           {
              path = path.substring(2);
              if (path.startsWith("["))
              { //indication of which child
                 idx = Integer.parseInt(path.substring(path.indexOf("[")+1, 
                                       path.indexOf("]")));
                 parent = (ThGroup)parent.getChildren().get(idx-1);
                 path = path.substring(path.indexOf("]")+1);
              }
              else //by default, first child in the group
                 parent = (ThGroup)parent.getChildren().get(0);
           }
        }
     }//end else
     
     //find current StructMapPage
     if (((String)prop.get("SpaceMapPage")).equals("ON"))
     {
        currpnum = findParentPage(sys, parent.getId());
        if (currpnum < 0)
           currpnum = rootPageNum;
     }
    
     boolean natixlog = true;    
     //insert the node
     if (insertgroup)
        insertGroup(txid, parent, currpnum, pat, contents, sys, natixlog, islog);
     else 
        insertLeave(txid, parent, currpnum, pat, contents, sys, natixlog, islog);
    
     return (String)prop.get("EC0");
  }  
  
  public String insertGroup(String txid, ThGroup parent, int pnum,
                            String path, Object contents, ThSystem sys,
                            boolean natixlog, boolean islog)
  { 
     //declaration & initialization           
     String attr, 
           data;
     Properties prop = sys.getProperties();        
     ArrayList values = (ArrayList)contents;
     String line = (String) values.remove(0);
     int idx = line.indexOf("=");
     if (idx >= 0)
     {
        attr = new String(line.substring(line.indexOf(" ")+1,idx));
        data = new String(line.substring(idx+1, line.length()));
     }
     else
     {
        idx = line.indexOf(" ");
        attr = new String(line.substring(0,idx));
        data = new String(line.substring(idx+1, line.length()));
     }    
       
     ThGroup group = storeGroupData(txid, path, attr, data, 
                               namE, sys, natixlog, LOG_YES);    
     
     if (((String)prop.get("SpaceMapPage")).equals("ON"))
     {
        IThPage page = sys.getBM().getPage(Integer.toString(pnum), sys);
        int datalength = ThStructMapPage.computeGRecLen();
        
        if (page.isFull(datalength))
        {
           page.add(group);
           pnum = splitPage(page, parent, group.getId(), sys);
        }
        else
        {
           page.add(group);
           page.updateFreeSpace(datalength);
        }
     }
    
     if (values.size() > 0)
        insertLeave(txid, group, pnum, path+"/L", values,
                    sys, natixlog, LOG_YES); 
    
     parent.addChild(group);
     
     return (String)prop.get("EC0");
  }
  
  public String insertLeave(String txid, ThGroup parent, int pnum,
                            String path, Object contents, ThSystem sys,
                            boolean natixlog, boolean islog)
  { 
     //declaration & initialization
     String pat = path;
     String line,
           attr,
           data;
     ThLeave leave;
     int idx;
     Properties prop = sys.getProperties();        
     ArrayList values = (ArrayList)contents;
   
     for(int i=0;i<values.size();i++)
     {
        data = (String)values.get(i); 
        leave = storeLeaveData(txid, pat, data, 
                        namE, sys, natixlog, islog);   
        parent.addChild(leave);
        
        if (((String)prop.get("SpaceMapPage")).equals("ON"))
        {
           IThPage page = sys.getBM().getPage(Integer.toString(pnum), sys);
           int datalength = ThStructMapPage.computeLRecLen();
           
           if (page.isFull(datalength))
           {
              page.add(leave);
              pnum = splitPage(page, parent, leave.getId(), sys);
           }
           else
           {
              page.add(leave);
              page.updateFreeSpace(datalength);
           }        
        }        
     }   
     return (String)prop.get("EC0");
  }
  
  public int splitPage(IThPage page, ThGroup subtree, int id, ThSystem sys)
  {
     IThPage rpage = (sys.getBM()).allocateStrucPage(sys);
     pageSet.push(Integer.toString(rpage.getPageNum()));
     
     page.removeSubTree(subtree);     
     rpage.addSubTree(subtree);
     
     IThNode rprox = new ThProxyNode(subtree.getId(),rpage.getPageNum());     
     page.add(rprox);
     page.updateFreeSpace(ThStructMapPage.computePRecLen());
     
     return rpage.getPageNum();
  }
    
 /**
  Handle INSERT operation
  @Returns: Return Code 
  @Receive: Transaction id,
             Path of the insert record,
             Leaves to be inserted,
             ThSystem
  */
  public synchronized String insertLoad(String txid, String path, 
                   Object leaves, ThSystem sys, boolean islog)
  { 
     sys.out("Inserting into XMLDoc...");
     String pat = path;
    
     Properties prop = sys.getProperties();        
     IThNode currPtr = (ThGroup)root.children.get(0);  //current position
     while (!path.equals(""))
     {
        if (path.equals("/"))
        {
        }
        if (path.startsWith("/G"))
        {
           //traversing the path
           path = path.substring(2);        
           if (path.startsWith("["))
           { //indication of which child
              int idx = Integer.parseInt(path.substring(path.indexOf("[")+1, 
                                     path.indexOf("]")));
              currPtr = (IThNode)currPtr.getChild(idx-1);
              path = path.substring(path.indexOf("]")+1);
           }
           else //by default, first child in the group
              currPtr = (IThNode)currPtr.getChild(0);
        }
        else if(path.startsWith("/L"))          
           path = path.substring(2);
      }
    
     //prepare node to add into this path
     ArrayList values = (ArrayList)leaves;
     boolean natixlog = false; // because the log records already on disk
     for(int i=0;i<values.size();i++)
     {
        String data = (String)values.get(i); 
        ThLeave leave = storeLeaveData(txid, pat, data, 
                        namE, sys, natixlog, islog);   
        ((ThGroup)currPtr).children.add(leave);
     }   
     return (String)prop.get("EC0");
  }
  
  /**
  * @Purpose: Handle LOAD operation.
  * @Return:  Return Code
  * @Receive: Transaction id,
  *           File handler,
  *           The name of xml document,
  *           ThSystem.
  */  
  public synchronized String load(String txid, DataInputStream in,
                                  String xmlname, ThSystem sys)
  {
     Stack stack = new Stack();
     Stack tmpStk = new Stack();
     String rc; //return code
     String path = "";
     boolean natixlog = true;
     stack.push(root);
     Properties prop = sys.getProperties();
    
     sys.out("Loading into XMLDoc...");
     ThBufferMgr BM = sys.getBM();
     BM.freeAllPages(sys);
     IThPage rootStrucPage = BM.getPage(Integer.toString(rootPageNum), sys);
 
     try
     {
        String inLine;    /*read in line by line*/
        String data;
        String attr;
        int id;
        boolean LOG_YES = true;
    
        while((inLine = in.readLine())!=null)
        {  
        /* If it reads in form <L> ...data...</L>, parse the data 
        *       between the 2 tags and creates leave node.
        * else if reads in only <L> ... data, parse data, push on stack
        * else if reads in only </L>, pop tage, <L>, pop data
        *                                    and create leave node
        * else if reads in only </L>, pop tag and it's not <L>, ERROR
        */       
           if (inLine.indexOf("<L") >= 0 && inLine.indexOf("</L>") >= 0)
           { //<L> ... </L>
              data = new String(inLine.substring(inLine.indexOf("<L")+3, 
                                           inLine.indexOf("</L>")));
        
              ThLeave leave = storeLeaveData(txid, path+"/L", data, 
                               xmlname, sys, natixlog, LOG_YES);     
              if (((String)prop.get("SpaceMapPage")).equals("ON "))
                 insertStruct (leave, sys, rootStrucPage, stack, strucPages); 
              ThGroup group = (ThGroup)stack.pop();
              group.children.add(leave);
              stack.push(group);
           }
           else if(inLine.indexOf("<L") >= 0 && inLine.indexOf("</L>") < 0)
           { //<L> ...
              tmpStk.push(inLine);
           }
           else if(inLine.indexOf("</L>") >= 0 && inLine.indexOf("<L") < 0)
           { //</L>
              data = (String)tmpStk.pop();
              if (data.indexOf("<L")<0)
              {  //<L> not found
                 return (String)prop.get("MISS_MATCH_LEAVE");
              }
              data = new String(data.substring(data.indexOf("<L")+3));
              ThLeave leave = storeLeaveData(txid, path+"/L", data, 
                          xmlname, sys, natixlog, LOG_YES); 
              if (((String)prop.get("SpaceMapPage")).equals("ON "))
                 insertStruct (leave, sys, rootStrucPage, stack, strucPages);
              ThGroup group = (ThGroup)stack.pop();
              group.children.add(leave);
              stack.push(group);
          }
          else if (inLine.indexOf("<G") >= 0)
          {/* If found <G, push on stack
            * else if found </G>, pop stack, construct G node
            */
             path += "/G";
             attr = new String(inLine.substring(inLine.indexOf("<G")+3, 
                                           inLine.indexOf("=")));
             data = new String(inLine.substring(inLine.indexOf("=")+2, 
                                           inLine.indexOf(">")-1));
                  
             ThGroup group = storeGroupData(txid, path, attr, data, 
                               xmlname, sys, natixlog, LOG_YES);   
             if (((String)prop.get("SpaceMapPage")).equals("ON "))
                insertStruct (group, sys, rootStrucPage, stack, strucPages);
             stack.push(group);
          }
          else if (inLine.indexOf("<XML") >= 0)
          {
             /* If found <XML, push on stack
             * else if found </XML>, pop stack, construct G node
             */          
             attr = new String(inLine.substring(inLine.indexOf(" ")+1, 
                                           inLine.indexOf("=")));
             data = new String(inLine.substring(inLine.indexOf("=")+2, 
                                           inLine.indexOf(">")-1));
       
             ThGroup group = storeGroupData(txid, ".", attr, data,  //"/"
                               xmlname, sys, natixlog, LOG_YES);
             if (((String)prop.get("SpaceMapPage")).equals("ON "))
                rootStrucPage.insert(group, -1, sys); // no parent. This is root  
             stack.push(group);
          }
          else if(inLine.indexOf("</G>") >= 0 || inLine.indexOf("</XML>") >= 0)
         {
            if (path.length() >= 2)
               path = path.substring(0, path.length() -2);
            ThGroup group = (ThGroup)stack.pop();    
            ThGroup parent = (ThGroup)stack.pop();
            parent.children.add(group); //add this Group node into tree
            stack.push(parent);
         }
      }     
      rc = (String)prop.get("EC0");
     }
     catch (Exception e)
     {
        e.printStackTrace();
        rc = (String)prop.get("EC3");
     }    
     return rc; 
  }
  
  /**
  * @Purpose: insert a node into the xml structure 
  *           stores in high level space map page.
  * @Return:  the page of the inserted data
  * @Receive: The node to be inserted,
  *           ThSystem,
  *           The root page,
  *           Stack,
  *           list of pages insert.
  */  
  public IThPage insertStruct(IThNode node, ThSystem sys, 
              IThPage rootStrucPage, Stack stack, ArrayList pages) 
  { 
     Properties prop = sys.getProperties();
     IThPage tmppage = null; 
     IThNode parentnode = (IThNode) stack.pop();
     IThPage newpage = null;
     IThNode temp = null, 
            rnode = null,
            proxy;
     int parentid = -1;
     String pnum;
     int id;
     IThNode prev = null;    
    
     ThStructMapPage rootpage = (ThStructMapPage) rootStrucPage;
     IThNode rootnode = rootpage.getNodeByOffset(0);
    
     if (parentnode != null) 
     {
        parentid = parentnode.getId(); 
        // 1. Find the page that contains the parent id
        // Try the current page first. 
        // If getIdx return -1, do a full search, starting from root.
        tmppage = findPageToInsert(parentid, pages, sys);
              
        // 2. check if the page is full. 
        // If not, insert. Otherwise, split page
        int reclength = node.size()+8;
        if (tmppage != null)
        {
           if (!tmppage.isFull(reclength)) 
              tmppage.insert(node, parentid, sys);
           else 
           { // handle page overflow, split page
              sys.out("In Split Page");
              newpage = (sys.getBM()).allocateStrucPage(sys); 
              String pagenum = Integer.toString(newpage.getPageNum());
          
              pages.add(pagenum);
              //tmppage is currently overflowed.
              tmppage.insert(node, parentid, sys);
         
              tmppage.split(newpage, parentid, stack, sys);      
           }
        }
        stack.push(parentnode);
    }    
    return tmppage;
  }
  
  /**
  * @Purpose: Find the page with available space to insert
  *           a struct node.
  * @Return: The page to insert
  * @Receive: parent id,
  *           array of pages,
  *           ThSystem.
  */  
  public IThPage findPageToInsert(int parentid, ArrayList pages, ThSystem sys)
  {
     IThNode node;
     IThPage page;
     String pnum; 
     int rc = -1;
     ThBufferMgr BM = sys.getBM();
    
     for (int i=0; i<pages.size(); i++)
     {
        pnum = (String) pages.get(i);
        page = (IThPage) BM.getPage(pnum, sys);
        rc = page.getOffset(parentid);
        if (rc >= 0)
          return page;
     }
    
     return null;
  }
 
  public int findParentPage(ThSystem sys, int parentid)
  {
     IThPage page = null;
     String pnum = "";
     ThBufferMgr BM = sys.getBM();
     while (!pageSet.empty())
     {
        pnum = (String) pageSet.pop();
        page = BM.getPage(pnum, sys);
        if (page.getOffset(parentid) >= 0)
           return Integer.parseInt(pnum);
     }
         
     return -1;
  }
  
  /**
  * @Purpose: Find the page with available space to insert
  *           a struct node.
  * @Return: The page to insert
  * @Receive: parent id,
  *           array of pages,
  *           ThSystem.
  */  
  public IThPage findPage(int parentid, ArrayList pagenums, 
                    IThNode node, ThSystem sys, 
                    IThPage rootStrucPage, 
                    IThPage currStrucPage, Stack stack)
  {
     IThPage parentpage = currStrucPage;
     int idx = currStrucPage.getOffset(parentid);
     IThNode nod;
     String pnum;
    
     if ( idx != -1) 
     {
        nod = currStrucPage.getNodeByOffset(idx);
        if (nod.getClass().getName().indexOf("ThProxyNode") < 0)
           parentpage = currStrucPage;
        else  // get page number from proxy node
        {
           pnum = Integer.toString(((ThProxyNode)node).getPageNum());
           parentpage = sys.getBM().getPage(pnum, sys);
        }
     } 
     else
     { //full search to find the page that contain parent node
        for (int i=0; i<pagenums.size(); i++)
        {
           pnum = (String)pagenums.get(i);
           parentpage = sys.getBM().getPage(pnum, sys);
           idx = parentpage.getOffset(parentid);
           if ( idx != -1)
           {
              nod = parentpage.getNodeByOffset(idx);
              if (nod.getClass().getName().indexOf("ThProxyNode") < 0)
                 currStrucPage = parentpage;
              else  // get page number from proxy node
              {
                 pnum = Integer.toString(((ThProxyNode)node).getPageNum());
                 parentpage = sys.getBM().getPage(pnum, sys);
              } // end else
            } // end if
         } // end for
      } // end else    
      return parentpage;
  }
    
  /**
  * @Purpose: Handle UPDATE operation
  * @Return:  Return Code
  * @Receive: Transaction id,
  *           name of the document,
  *           the path of inserted record,
  *           new value,
  *           ThSystem.
  */  
  public String update(String txid, String xmlname,  String path, 
                String newVal, ThSystem sys, boolean islog)
  {
     sys.out("Updating "+xmlname+"...");
    
     Properties prop = sys.getProperties();
     IThNode currPtr = (ThGroup) root.children.get(0);  //current position
     Stack stack = new Stack();
     int idx = 0;   
     ThBufferMgr BM = sys.getBM();
     int oldid;
     String oldval;
     ArrayList children;
     String rc = "";
     String pat = path;
    
     while (true)
     {
        if (path.startsWith("/G"))
        {
           path = path.substring(2);        
           stack.push(currPtr);
           if (path.startsWith("["))
           { //indication of which child
              idx = Integer.parseInt(path.substring(path.indexOf("[")+1, 
                                 path.indexOf("]")));
              idx -= 1; // array idx starts from 0
              currPtr = (IThNode)currPtr.getChild(idx);
              path = path.substring(path.indexOf("]")+1);
           }
           else //by default, first child in the group
              currPtr = (IThNode)currPtr.getChild(0);
        
           if (path.equals(""))
           {     //need to be done also
              currPtr = (ThGroup)stack.pop();
              break;        
           }      
        
        }
        else if (path.startsWith("/L"))
        {
           path = path.substring(2);        
        
           if (path.startsWith("["))
           { //indication of which child
              idx = Integer.parseInt(path.substring(path.indexOf("[")+1, 
                                 path.indexOf("]")));
              idx -= 1; // array idx starts from 0          
              children = ((ThGroup)currPtr).getChildren();
              if (children.size() <= idx)
                 return (String) prop.get("PATH_NOT_FOUND");
              ThLeave leave = (ThLeave)children.get(idx);
              oldid = leave.getId();      
              //new leave added
              rc = storeUpdateData(txid, oldid, leave, pat, newVal, sys, islog);              
           }
           else
           { //by default, first child in the group
              children = ((ThGroup)currPtr).getChildren();
              ThLeave leave = (ThLeave)children.get(0);
              oldid = leave.getId();
              //new leave added
              rc = storeUpdateData(txid, oldid, leave, pat, newVal, sys, islog);
           }
           break;  
       }
       else
       {
          sys.out("something wrong!");
          break;
       }
     }  
     return rc;
  }  
  
  /**
  * @Purpose: This function is called by ROLLBACK
  *         to perform undo for an operation.
  * @Return:  Return Code.
  * @Receive: Log record for undo operation, 
  *           ThSystem.
  */  
  public int undo(ThLogRecord log, ThSystem sys)
  {
     String logtype = log.getLogType();
     if (logtype.indexOf("UINSERT") >= 0)
        undoInsert(log.getId(), (ThGroup)root);
     else if(logtype.equals("UDELETE"))
        undoDeleteGroup(log.getPath(), (ThGroup)root.children.get(0), sys);
     else if(logtype.equals("UUPDATE"))
        undoUpdate(log, sys);      
     else
        return -1; //unknown log record    
     return 0; // success
  }
  
 /**
  * @Purpose: This function is called by RESTART RECOVERY
  *         to perform redo for an operation.
  * @Return:  Return Code.
  * @Receive: Log record for redo operation, and ThSystem.
  */  
  public int redo(ThLogRecord log, ThSystem sys)
  {
     String logtype = log.getLogType();
    
     if (logtype.indexOf("UINSERT") >= 0) 
        redoInsert(log, sys);
     else if(logtype.equals("UDELETE"))
        redoDeleteGroup(log, sys);
     else if(logtype.equals("UUPDATE"))
        redoUpdate(log, sys);      
     else
        return -1;    
       
     return 0;
  }  
  
  /**
  * @Purpose: Perform redo insert
  * @Return:  Return Code
  * @Receive: Log record, and ThSystem
  */  
  private int redoInsert(ThLogRecord log, ThSystem sys)
  {
     String id = log.getId();
     boolean exist = find(id, sys);
     boolean natixlog = false;
    
     if (exist)
       return 0;
    
     String rc = insert(log.getTxId(), log.getPath(), 
             log.getNewValues(), sys, LOG_NO);
     return Integer.parseInt(rc);
  }
  
  /**
  * @Purpose: Find to see if a record id is exist in the tree.
  * @Return:  True if a record id is already in the tree.
  *           False otherwise.
  * @Receive: id to be search, and ThSystem
  */  
  public boolean find(String id, ThSystem sys)
  {    
     if (((ThGroup)root).getChildrenCount() < 1)
        return false;
    
     if ((((ThGroup)root.children.get(0)).getChildrenCount()) < 1)
        return false;
    
     return findInNode(id, (ThGroup)root.children.get(0), sys);
  }
  
 /**
  * @Purpose: Find an id in a node
  * @Return:  True if id exist,
  *           False otherwise.
  * @Receive: id to be search, the node, and ThSystem.
  */
  public boolean findInNode(String id, IThNode node, ThSystem sys)
  {  
     if (node.isMatch(id))
        return true;
    
     if (node.getClass().getName().indexOf("ThGroup") >= 0)
     {
        ThGroup group = (ThGroup) node;
        for (int i=0;i<group.children.size();i++)
        {
           AThNode obj = (AThNode)group.children.get(i);
      
           if (obj.isMatch(id))
           return true;
        }
     }
    
     return false;
  }
  
  /**
  Redo of a delete operation
  @Return:  Recturn Code.
  @Receive: Log record, and ThSystem.
  */  
  private int redoDeleteGroup(ThLogRecord log, ThSystem sys)
  {
     ArrayList path = new ArrayList();
     path.add(log.getPath());
     delete(log.getTxId(), path, sys, LOG_NO);
    
     return 0;
  }
  
  /**
  * @Purpose: Redo of an update operation
  * @Return: Return Code.
  * @Receive: Log record, and ThSystem.
  */  
  private int redoUpdate(ThLogRecord log, ThSystem sys)
  {
     update(log.getTxId(), log.getXmlName(),  log.getPath(), 
           log.getNewValue(), sys, LOG_NO);
     return 0;
  }
  
 /**
  * @Purpose: Undo of an update operation
  * @Return:  nothing
  * @Receive: id of the updating record, and root of the tree.
  */
  private void undoInsert(String id, IThNode root)
  {
     int theid = Integer.parseInt(id);
     IThNode child;
     IThNode parent = root;
   
     if (root.getChildrenCount() < 1)
        return;
    
     ArrayList children = (ArrayList)parent.getChildren();
     int num = children.size();
      
     for (int i=0;i<num;i++)
     {
        child = (IThNode) children.get(i); 
        if (child.getId() == theid)
        {
           children.remove(i);
           return;
        }
        else
          undoInsert(id, child);
     } 
  }
  
  /**
  * @Purpose: Undo of a delete operation
  * @Return:  Nothing
  * @Receive: The path of deleting record, and parent node.
  */ 
  private void undoDeleteGroup(String path, ThGroup parent, ThSystem sys)
  {    
     Stack stack = new Stack();
     int idx = 0;
     IThNode currPtr = parent;
     String pnum;
     sys.out("...Undo Delete");
     ArrayList children;
     while (true)
     {   //travese down the tree until the condition is matched
        if (path.startsWith("/G"))
        {
           path = path.substring(2);        
           stack.push(currPtr);
           if (path.startsWith("["))
           { //indication of which child
              idx = Integer.parseInt(path.substring(path.indexOf("[")+1, 
                                 path.indexOf("]")));
              idx -= 1; // array idx starts from 0
              currPtr = (IThNode)currPtr.getChild(idx);
              path = path.substring(path.indexOf("]")+1);
           }
           else
           { //by default, first child in the group
              currPtr = (IThNode)currPtr.getChild(0);
           }
        
           if (path.equals(""))
           { //default, 1st child in the list of this group 
              currPtr = (ThGroup)stack.pop();
              ThAttributeNode attrnode = ((ThGroup)currPtr).getAttributeNode();
              children = ((ThGroup)currPtr).getChildren();

              ((IThNode)children.get(idx)).setDeleted(false);
               break;        
           }      
        }
        else if (path.startsWith("/L"))
        {
           path = path.substring(2);        
           currPtr = (ThGroup)stack.pop();
        
           if (path.startsWith("["))
           { //indication of which child
              idx = Integer.parseInt(path.substring(path.indexOf("[")+1, 
                                 path.indexOf("]")));
              idx -= 1; // array idx starts from 0
          
              children = ((ThGroup)currPtr).getChildren();
              ThLeave leave = (ThLeave)children.get(idx);
              ((IThNode)children.get(idx)).setDeleted(false);
           }
           else
           { //by default, first child in the group
              children = ((ThGroup)currPtr).getChildren();
              ThLeave leave = (ThLeave)children.get(0);
            
             ((IThNode)children.get(0)).setDeleted(false);
           }
           break;  
        }
        else
        {
           sys.out("something wrong!");
           break;
        }
     } 
  }
  
  /**
  * @Purpose: Undo of an update operation
  * @Return:  Nothing
  * @Receive: Log record
  */  
  private void undoUpdate(ThLogRecord log, ThSystem sys)
  {
     sys.out("Undo Update......");
     String pnum = log.getPageNum();
     ThBufferMgr BM = sys.getBM();
     IThPage page = BM.getPage(pnum, sys); 
     int id = Integer.parseInt(log.getId());
     String oldval = log.getOldValue();
     int len = oldval.length();
     String lsn = log.getLSN();
     page.update(id, oldval, lsn, sys); 
  } 
         
  /**
  * @Purpose: First, compute the space needed for a 
  *             a record to be inserted into a page.
  *           Next, allocate the page for inserted.
  *           Next, store data in the page.
  * @Return:  The logical leave node.
  * @Receive: Transaction id of this operation,
  *           The page to be inserted,
  *           The insert value,
  *           The name of the xml document,
  */  
  private ThLeave storeLeaveData(String txid, String path, 
         String value, String xmlname, ThSystem sys, 
         boolean natixlog, boolean islog)
  {
     //1. compute the length needed for this data to be on data page
     int datalength = ThPage.computeLRecLen(value); //header for a leave node
     int id = ThSystem.getId();   //unique logical id for this leave node
     //2. allocate page to store data
     IThPage page = sys.getBM().getPage(datalength, sys); 
     //3. store data in the page
     String msg = "UINSERTLEAVE";
     ArrayList data = new ArrayList();
     data.add(value);
     page.insert(msg, txid, id, path, data, xmlname,
           sys.getDB(xmlname), false, sys, natixlog, islog);
     page.updateFreeSpace(datalength);
     
     return new ThLeave(id, page.getPageNum(), sys);
  }
  
  /**
  * @Purpose: First, compute the space needed for a 
  *             a record to be inserted into a page.
  *           Next, allocate the page for inserted.
  *           Next, store data in the page.
  * @Return:  The logical leave node.
  * @Receive: Transaction id of this operation,
  *           The page to be inserted,
  *           The insert value,
  *           The name of the xml document,
  */  
  private ThGroup storeGroupData(String txid, String path, 
                  String attr1, String attr2, String xmlname, 
                  ThSystem sys, boolean natixlog, boolean islog)
  {
     //1. compute the length needed for this data to be on data page
     int datalength = ThPage.computeGRecLen(attr1, attr2);
     //(attr1.length()+attr2.length())*2 + 20; //header for a leave node in data page
     int id = ThSystem.getId();   //unique logical id
     //2. allocate page to store data    
     IThPage page = sys.getBM().getPage(datalength, sys);
     //3. store data in the page
     String msg = "UINSERTGROUP";
     ArrayList data = new ArrayList();
     data.add(attr1);
     data.add(attr2);
     page.insert(msg, txid, id, path, data, xmlname,
           sys.getDB(xmlname), true, sys, natixlog, islog);   
     page.updateFreeSpace(datalength);
     
     return new ThGroup(id, attr1, attr2, page.getPageNum());
  }
  
  /**
  * @Purpose: Get the page that contains the update record.
  * @Return:  Return code
  * @Receive: Transaction id of this operation,
  *           Id of the updating data,
  *           The leave node
  *           The pathm
  *           New data
  */  
  private String storeUpdateData(String txid, int id, ThLeave leave, 
              String path, String newVal, ThSystem sys, boolean islog)
  {
     int datlength = newVal.length();
     ThBufferMgr BM = sys.getBM();
     //locate the page
     String pnum = Integer.toString(leave.getPageNum());
     IThPage page = BM.getPage(pnum, sys);
     //create log record
     String oldval = (String) (page.getData(id)).get(0);
     String msg = "UUPDATE";
     String lsn = ""; 
     if (islog)  
     {
        lsn = page.createUpdateLogRecord(msg, txid, id, path, 
                      oldval, newVal, namE, sys.getDB(namE), sys);
     }
     String rc = page.update(id, newVal, lsn, sys); 
    
     return "0"; // return rc;
  }  
  
  /**
  * @Purpose: Free the Strucpage
  * @Return:  nothing
  * @Receive: ThSystem where global variables held.
  */  
  public void freeStrucPages(ThSystem sys)
  {
     root = null;
     int iD = ThSystem.getId();
     root = new ThGroup(iD, "ROOT", "ROOT");
     rootPageNum = ((sys.getBM()).allocateStrucPage(sys)).getPageNum();
  }
   
  /**
  * @Purpose: Print the tree structure and its data
  * @Return:  nothing
  * @Receive: Transaction id, 
  *           ThSystem.
  */  
  public void printTree(String txid, ThSystem sys)
  {
     sys.out("Printing structure of XML Tree:  "+namE);
    
     if (((ThGroup)root).getChildrenCount() < 1)
     {
        sys.out("Empty tree!");
        return;
     }
     printNode((ThGroup)root.children.get(0), sys);
  }
  
  /**
  * @Purpose: Print data in a node.
  * @Return:  Nothing
  * @Receive: The node.
  */  
  public void printNode(ThGroup node, ThSystem sys)
  {  
     sys.out(spaces+" "+node.toString(sys));
     spaces += "  ";
     String data = "";
     
     ArrayList children = node.getChildren();    
     for (int i=0;i<children.size();i++)
     {
        AThNode obj = (AThNode)children.get(i);
      
        if (obj.isDeleted())
           continue;      
 
        if (((obj.getClass()).getName()).indexOf("ThGroup") >= 0)
        { 
           printNode((ThGroup)obj, sys);
        }
        else
        {     //it's a leave node
           data = ((ThLeave)obj).toString(sys);
           sys.out(spaces +" "+data);
        }
     }
     spaces = spaces.substring(2);
  }
  
  /**
  * @Purpose: 
  * @Return:   
  * @Receive:  
  */  
  public void printNodeToResultSet(ThGroup node, ArrayList arr, ThSystem sys)
  {
     arr.add(spaces+node.attribute.toString());
     spaces += "  ";
     String data = "";
    
     for (int i=0;i<node.children.size();i++)
     {
        Object obj = node.children.get(i);
 
        if (((obj.getClass()).getName()).indexOf("ThGroup") >= 0)
        {
           printNode((ThGroup)obj, sys);
        }
        else
        {     //it's a leave node
           arr.add(spaces + data);
        }
     }
     spaces = spaces.substring(2); 
  }    
  
  public void printSpaceMapPages(ThSystem sys)
  {
     sys.out("Printing SpaceMapPages: ");
     ThStructMapPage page = (ThStructMapPage) sys.getBM().getPage(Integer.toString(rootPageNum), sys);
     IThNode root = page.getNodeByOffset(0);
     printStructMapNode(root, page, sys);
  }
  
  public void printStructMapNode(IThNode node, ThStructMapPage page, ThSystem sys)
  {
  }
  
  public String getName() 
  {
     return namE;
  }
  
  public ThGroup getRoot()
  {
     return root;
  }
}