package cs297.xmldb.src.server;

import java.io.*;
import java.util.*;
import cs297.xmldb.src.cm.*;
import cs297.xmldb.src.sm.*;
import cs297.xmldb.src.bm.*;

/**
 This class creates an instans of a specific database when it is called.
 It implements Serializable because we need
 to simulate image copy of the database when archive is specified.
 @Date: Spring & Fall 2004
 @author: Thien An Nguyen, 
 */
public class ThDB implements Serializable
{
   boolean dbUp = false; //no access to this db
   Hashtable docNames;
   String name = null;

  public ThDB(String dbName)
  {
     name = new String(dbName);
     docNames = new Hashtable();
     dbUp = true;
  }
  
  /**
   When database is up, access to db is allow. 
   If dbUp = false, no access to objects in the database.
   @Receive: nothing
   @Return:  nothing
   */
  public void startDB()
  {
     dbUp = true;
  }
  
  /**
   When dbUp is false, access to db is denied.
   */
  public void stopDB()
  {
     dbUp = false;
  }
  
  public boolean insertDoc(String docName)
  {
     return false;
  }
  
  public boolean deleteDoc(String docName)
  {
     return false;
  }
  
  /**
   This function check to see if the name of an xml
   document is already exist in the database. If it
   is already in the database, return a message. If
   not, create the document in this database, get 
   an image copy of this database for future rollbacks and recovery.
   @Receive: transaction id,
             name of the document,
             ThSystem
   @Returns: return code
   */
  public String createXmlDocEntry(String txid, String xmlname, ThSystem sys)
  {    
     Properties prop = sys.getProperties();
   
     if (docNames.contains(xmlname))
        return (String)prop.get("DOC_EXISTED");

     /* create xmldoc object and have an image copy of it */
     docNames.put(xmlname, new ThXMLDoc(xmlname, sys));
     ThUtilities.imageCopy(this, sys);
     sys.out(xmlname+" added in "+name);
   
     return (String)prop.get("EC0");
 }
 
  /**
   Remove a document from this database
   @Receive: name of the document
   @Returns: Return Code
   */
  public String dropXmlDoc(String xmlname, ThSystem sys)
  { 
     Properties prop = sys.getProperties();
     if (!docNames.contains("xmlname"))
        return (String)prop.get("DOC_NOT_FOUND");
   
     docNames.remove(xmlname);
     sys.out(xmlname + " dropped");    
     
     return (String)prop.get("EC0");      
  }  
 
  public ThXMLDoc getDocument(String name)
  {
     return (ThXMLDoc)docNames.get(name);
  }
 
  public Hashtable getDocNames()
  {
     return docNames;
  }
 
  public String getName()
  {
     return name;
  }
 
  public boolean contains(String xmlname)
  {
     return docNames.containsKey(xmlname);
  }
}