package cs297.xmldb.src.server;

import java.io.*;
import java.util.*;
import cs297.xmldb.src.sm.*; 
import cs297.xmldb.src.cm.*;
import cs297.xmldb.src.bm.*;


/**
 This class contains global variables such as system
 catalog (where to search for created databases,...
 @Date: Spring 2004
 @author: Thien An Nguyen, 
 */
public class ThSystem implements Serializable{
  static Properties propertieS = new Properties();
  boolean sysUp = false; 
  Hashtable cataLog;
  ThBufferMgr BM;
  ThLogManager LM; 
  ThStorageMgr SM;
  ThTransactionMgr TM;
  ThRecoveryMgr RM;
  static int id = 1;
  int LSN;
  DataOutputStream outStream;
  DataOutputStream verifyStream;
  boolean NATIX = false;
  int timeStamp = 0;  
  boolean iO = true;
  
  /*
   This constructor load the properties file 
   and initialize global variables */
  public ThSystem()
  {
     try
     {     
        propertieS.load(new FileInputStream("c://an//cs297//xmldb//dbsystem.properties"));
      
        LSN = Integer.parseInt((String)propertieS.get("pageLSN"));
      
        String filename = (String)propertieS.get("outputFile");
        outStream = new DataOutputStream(
                      new BufferedOutputStream(
                       new FileOutputStream(filename)));  
 
        cataLog = new Hashtable();
        if (((String)propertieS.get("IO")).equals("YES"))
           iO = true;
        else
           iO = false;
     }
     catch(Exception e)
     {
        cataLog = new Hashtable();
     }  
    
     BM = new ThBufferMgr(propertieS);
     SM = new ThStorageMgr(propertieS); 
     LM = new ThLogManager(propertieS);
     TM = new ThTransactionMgr(propertieS);
    
     String rmethod = (String)propertieS.get("recoverMethode");
     if (rmethod.equals("NATIX"))
        NATIX = true;
    
     RM = new ThRecoveryMgr(propertieS);
     start();
  }
  
  /**
   This function checks if xmldoc name already in the database.
   If it is, add it in the list of database's name.
   If not, return error code for document not found
   @Params: the xmldocument name, 
            StringBuffer dbname
   @Returns: return code
   */
  public String findDbContainsXmlDoc(String xmlname, StringBuffer dbname)
  {
    for (Enumeration e = cataLog.elements(); e.hasMoreElements() ;) 
    {
       ThDB db = (ThDB)e.nextElement();
       if (db.contains(xmlname))
       {
          dbname.append(db.name);
          return (String)propertieS.get("EC0");     
       }
    }
   
    return (String)propertieS.get("DOC_NOT_FOUND");     
  }
 
  /**
   This function fist checks to see if the document is in any database 
   in the system. If the return code is 0, it gets to the database from 
   the catalog Hashtable and call load function from the xmldoc object.
   */ 
  public String loadIntoXmlDoc(String txid,String filename, String xmlname)
  {    
    StringBuffer dbname = new StringBuffer();
    String rc = findDbContainsXmlDoc(xmlname, dbname);
    if (rc.equals((String)propertieS.get("EC0")))
    {
       ThDB db = (ThDB) cataLog.get(dbname.toString());
       ThXMLDoc xmldoc = db.getDocument(xmlname); 
       return SM.loadXMLDoc(txid, xmldoc, filename, xmlname, this);
    }
      
    return rc;
  }
 
 /**
  This function will create an entry of database name
  in the catalog hash table
  @receive: the database name to be created
  @return: Error code if system is down or 
                      database already exist
            Success code if success.
  */
 public String createDB(String txid, String dbName)
 {  /*create an entry in catalog table for */
    if(!sysUp)               /* SYS_DOWN   */
       return (String)propertieS.get("SYS_DOWN");
  
    if (cataLog.contains(dbName))
       return (String)propertieS.get("DB_EXISTED"); 
     
    cataLog.put(dbName, new ThDB(dbName)); 
    out(dbName + " created");
   
    return (String)propertieS.get("EC0");
 }
 
 /**
  This function will drop a dabase.
  When user specified this command, objects in this 
  database will also be dropped.
  
  @receive: the database name to be dropped
  @return: Error code if errors.
           Success code if success.
  */
 public String dropDB(String txid, String dbname)
 {  /* remove this db from catalog table */
    if(!sysUp)
       return (String)propertieS.get("SYS_DOWN");
   
    if (!cataLog.containsKey(dbname))
    {
       out("Drop "+dbname+ " not found!");
       return (String)propertieS.get("DB_NOT_FOUND");
    }
       
    cataLog.remove(dbname);   
    out(dbname + " dropped");     
    return (String)propertieS.get("EC0");
 }
 
 /**
  This function creates and xml document in the database
  @params: transaction id,
           the document name
           the database's name that contains the document.
  */
 public String createXmlDocEntry(String txid, String xmlname, String dbname)
 {
    if (!cataLog.containsKey(dbname))
       return (String)propertieS.get("DB_NOT_FOUND");
   
    ThDB db = (ThDB)cataLog.get(dbname);
    return db.createXmlDocEntry(txid, xmlname, this);
 }
 
 /** 
  This function will traverse through the list of databases in 
  the catalog and call dropXmlDoc within each database to drop
  the xmldocname.
  If this drop operation success, it will exit out of the 
     while loop and return with return code 000
   If the drop from a specific database is fail, 
     it continues to the next database in the catalog.
   If it couldn't fin the xmlname, it returns error
  */
 public String dropXmlDoc(String txid,String xmlname)
 {
    String rc;
    for (Enumeration e = cataLog.elements(); e.hasMoreElements() ;) 
    {
       ThDB db = (ThDB)e.nextElement();
       rc = db.dropXmlDoc(xmlname, this);
     
       if (rc.equals(propertieS.get("EC0"))) //ok to exit now
          BM.flushAllPage(this);
         
      return rc;
   }
   
   return (String)propertieS.get("DOC_NOT_FOUND");
 }
 
 public static int getId()
 {
    return id++;
 }
 
  public boolean start()
  {    
    sysUp = true;
    systemOut("Sub-system is up");
    return true;
  }
  
  public boolean stop()
  {
     sysUp = false;   
     out("Sub-system is stopped");
     return true;
  } 
  
  public ThBufferMgr getBM()
  {
     return BM;
  }
  
  public ThLogManager getLM()
  {
     return LM;
  }
  
  public ThStorageMgr getSM()
  {
     return SM;
  }  
  
  public ThTransactionMgr getTM()
  {
     return TM;
  }
  
  public ThRecoveryMgr getRM()
  {
      return RM;
  }
  
  public Properties getProperties()
  {
     return propertieS;
  }
  
  public String getLSN()
  {
     String lsn = Integer.toString(LSN);
     LSN++;
     return lsn;
  }
  
  public String getCurrentLSN()
  {
     return Integer.toString(LSN);
  }
  
  public Hashtable getCatalog()
  {
     return cataLog;
  }
  /**
   This function returns the database's object from the catalog table.
   @params: the name of the document
   @return: the database's name
   */
  public String getDB(String xmlname)
  {
     for (Enumeration e = cataLog.elements(); e.hasMoreElements() ;) 
     { 
        ThDB db = (ThDB)e.nextElement();
        if (db.contains(xmlname))
           return db.getName();
     }
    
     return (String)propertieS.get("EC2");
  }
  
  public boolean isNatix()
  {
     return NATIX;
  }
  
  public int getCurrentTimeStamp()
  {
     timeStamp++;
     return timeStamp;
  }
  
  public int getLastTimeStamp()
  {
     return timeStamp;
  }
  
  /**
   This function is called when the system
   detects a crash (simulation). It first restores
   to the last image copy. Then it calls recovery 
   manager to perform 3 phases of recovery.
   */
  public String recoverFromCrash()
  {
     int rc;
     /* restore to last image copy of DBs in catalog */
     rc = restoreToLastImgCpy();
     rc = RM.restartRecovery(this);
     if (rc == 0)
       return (String) propertieS.get("EC0");
     
     return (Integer.toString(rc));
  }
  
  /**
   This function handle the restore of the last image copy.
   It will remove the current version of the databases in the system
   and restore the version of the databases from the last system
   checkpoint.
   */
  public int restoreToLastImgCpy()
  {
     String dbname;
     ThDB db;

     for (Enumeration e = cataLog.keys(); e.hasMoreElements();)
     {
        dbname = (String) e.nextElement();
        /* remove current version of DB */
        cataLog.remove(dbname);
        db = ThUtilities.getImageCopy(dbname, this);
        cataLog.put(dbname, db);
     }
    
     return 0;
  }
  
  public int checkPoint(String dbname)
  {
     TM.removeTransactions();
     return 0;
  }
  
  /**
   This if system IO funciton
   */
  public void out(String line)
  {
     try
     {
        if (iO)
           System.out.println(line);
        else
        {
           outStream.writeBytes(line+"\r\n");
           outStream.flush();
        }
     }
     catch (IOException e){
        e.printStackTrace();
     }
  }
  
  /**
   This if system IO funciton
   */  
  public void systemOut(String line)
  {
     try
     {
        if (iO)
           System.out.println(line);
        else
        {
           outStream.writeBytes(line+"\r\n");
           outStream.flush();
        }
     }
     catch (IOException e){
        e.printStackTrace();
     }
  } 
}