package cs297.xmldb.src.sm;
import java.io.*;
import java.util.*;
import cs297.xmldb.src.cm.*;
import cs297.xmldb.src.bm.*;
import cs297.xmldb.src.server.*;
import cs297.xmldb.src.sm.*;

/**
 COPYRIGHT (C) 2004 Thien An Nguyen. All Rights Reserved.
 This class contains global variables such as system
 catalog (where to search for created databases,...
 @Date: Spring 2004
 @author: Thien An Nguyen, 
 */
public class ThLogManager
{
  ArrayList logRecords;
  DataInputStream inStream;
  DataOutputStream outStream;
  
  public ThLogManager(Properties properties)
  {
     logRecords = new ArrayList();
     String filename = (String)properties.get("logRecords");
     try
     {
        outStream = new DataOutputStream(
                   new BufferedOutputStream(
                    new FileOutputStream(filename)));
    
        inStream = new DataInputStream(
                   new BufferedInputStream(
                    new FileInputStream(filename)));   
     }
     catch(Exception e)
     {
        e.toString();
     }       
  }
  
  /**
   write this Log manager object on disk
   @param: the properties object
   @return: the name of the file
   */
  public String flush(Properties properties)
  {  
     String outfile = properties.getProperty("logRecords");
    
     try
     {
        ObjectOutputStream outStream = 
           new ObjectOutputStream(new FileOutputStream(outfile));
        outStream.writeObject(this);
     }
     catch(Exception e)
     {
        e.toString();
        return null;
     }     
     return outfile;  /* return the name of the file*/
  }
  
  /**
   Open the the file that contains the log records
  */
  public int load(Properties properties)
  {  
     String infile = properties.getProperty("logRecords");
    
     try
     {
        ObjectInputStream inStream = 
           new ObjectInputStream(new FileInputStream(infile));
        inStream.readObject();
     }
     catch(Exception e)
     {
        e.toString();
        return -1;
     }     
     return 0;  //success
  }
  
  /**
   Format the log records and write them onto the disk.
   @params:  the log type,
             log LSN,
             the dirty page number,
             transaction id,
             offset of the node in the page,
             node id,
             the name of the xml object
             the name of the database contains the object,
             value before changed,
             value after the changed.
   @returns: false if IO error.
   */
  public boolean writeLogRecord(String logtype, String lsn, String pNum,  
                     String txid, int offset, String id, String xmlname, 
                     String dbname, String oldval, String newval)
  {
     try 
     {
        outStream.writeBytes(logtype+" "+lsn+" "+pNum+" "+txid
                 +" "+offset+" "+id+" "+xmlname+" "+dbname
                 +" "+oldval+" "+newval
                 +"\r\n");
        outStream.flush();
     }
     catch (Exception e)
     {
        e.toString();
        return false;
     }
    
     return true; //success
  }
  
  /**
   Write log record using the formated string passed in
   */
  public boolean writeLogRecord(String info)
  {
     try 
     {
        outStream.writeBytes(info+"\r\n");
        outStream.flush();
     }
     catch (Exception e)  /* if IO error, return false */
     {
        e.printStackTrace();
        return false;
     }
    
     return true; 
  }
  
  /**
   Call writeLogRecord to format and write a log record onto disk 
   @params:  the log type,
             log LSN,
             the dirty page number,
             transaction id,
             offset of the node in the page,
             node id,
             the name of the xml object
             the name of the database contains the object,
             value before changed,
             value after the changed.
   @returns: false if IO error.
   */
  public String createLogRecord(String logtype, String pnum, String txid, int offset, 
                                String oldVal, String newVal, String id, String lsn, 
                                String xmlname, String dbname)
  {
      writeLogRecord(logtype, lsn, pnum, txid, offset, id, xmlname, dbname,
                     oldVal, newVal);
      
      return lsn;
  }  
  
  private String removeStart(String original, String key)
  {
     if (original.startsWith(key))
        return original.substring(1);
    
     return original;
  }
      
  private String removeEnd(String original, String key)
  {
     if (original.endsWith(key))
        return original.substring(0, original.length()-1);
    
     return original;
  }
  
  /**
   Load a log record into memory from disk.
   @params: StringTokenizer contains the log record's text.
            System object,
            log type,
            log LSN
   */
  public ThLogRecord createLogRecord(StringTokenizer tk, 
                  ThSystem sys, String logtype, String lsn)
  {  
     String oldval = "",
            newval = "";
     String dbname = tk.nextToken();
     String xmlname = tk.nextToken();
     String pnum = tk.nextToken();        
     String txid = tk.nextToken();
     String id = tk.nextToken(); 
    
     if (!sys.isNatix() || logtype.equals("UUPDATE")
        || logtype.equals("UDELETE")) /* ARIES log records */
     {                
        if (logtype.indexOf("INSERT") >= 0)
        {
           String path = tk.nextToken();
           newval = getNewValForInsert(tk);
           
           return new ThLogRecord(logtype,lsn,pnum,txid,id,
                                  path,newval,xmlname,dbname);
        }
        else if (logtype.indexOf("UUPDATE") >= 0)
        {
           String path = tk.nextToken();           
           oldval = getOldValForUpdate(tk);           
           newval = getNewValForUpdate(tk);
           
           return new ThLogRecord(logtype,lsn,pnum,txid,id, 
                       path ,oldval, newval, xmlname, dbname); 
        }
        else if (logtype.indexOf("UDELETE") >= 0)
        {
           String path = tk.nextToken();

           return new ThLogRecord(logtype,lsn,pnum,txid,id, 
                                  path,"", xmlname, dbname); 
        }
     } 
     else            /* NATIX log records */
     {
        String path = tk.nextToken();        
        newval = getNewValForNatix(tk);
        ThLogRecord log = new ThLogRecord(logtype,lsn,pnum,txid,id,
                                        path,newval,xmlname,dbname);
        String leave;
        while (tk.hasMoreTokens())
        {
           leave = getLeaveVal(tk);
           log.addLeave(leave);
        }
        
        return log;
    }
    
    sys.out("Doesn't understand log record type!");
    return null;
  }
  
  /**
   This function parses the new value from a log record.
   @param: StringTokenizer contains this value
   @return: The new value of the log record
   */
  private String getNewValForInsert(StringTokenizer tk)
  {
     String newval = tk.nextToken();
     while (tk.hasMoreTokens())
        newval += " "+tk.nextToken();      

     newval = removeStart(newval, "(");
     newval = removeEnd(newval, ")");      
     
     return newval;
  }
  
  /**
   This function parses the old value from a log record.
   @param: StringTokenizer contains this value
   @return: The old value of the update log record
   */
  private String getOldValForUpdate(StringTokenizer tk)
  {
     String oldval = tk.nextToken();
     while (tk.hasMoreTokens())
     {
        oldval += " "+tk.nextToken();
        if (oldval.indexOf(")") >= 0)
           return oldval.substring(1, oldval.length()-1);        
     } 
     
     return oldval;      
  }
  
 /**
   This function parses the old value from a log record.
   @param: StringTokenizer contains this value
   @return: The old value of the update log record
   */
  private String getNewValForUpdate(StringTokenizer tk)
  {
      String newval = tk.nextToken();
      while (tk.hasMoreTokens())
      {
         newval += " "+tk.nextToken();
         if (newval.indexOf(")") >= 0)
            return newval.substring(1, newval.length()-1);
         }
       
      return newval;
  }
  
  /**
   This function parses the new value from a log record
   that used Natix's log record format.
   @param: StringTokenizer contains this value
   @return: The new value of a Natix log record
   */
  private String getNewValForNatix(StringTokenizer tk)
  {
     String newval = tk.nextToken();
     while (newval.indexOf(")") < 0) 
        newval += " "+tk.nextToken();   
     
     return newval.substring(1, newval.length()-1);
  }
  
  /**
   This function parses the leave value of a leave from a log record.
   @param: StringTokenizer contains this value
   @return: The leave value.
   */    
  private String getLeaveVal(StringTokenizer tk)
  {
      String leave = tk.nextToken(); 
      while (tk.hasMoreTokens())
      {
         leave += " "+tk.nextToken(); 
         if (leave.indexOf(')') >=0)
            return leave.substring(1, leave.length()-1);
      }
            
      return leave;
  }
    
  /**
   This function load the log records from disk using the list
   of lsn of the log records.
   @params: The list of LSNs of the records that need to be loaded
            System object
   @return: ArrayList of the log records.
   */  
  public ArrayList getLogRecords(ArrayList lsnlist, ThSystem sys)
  {
     String inline;
     ArrayList logs = new ArrayList();
     StringTokenizer tk;
     String logtype;
     String lsn;               /* lsn of the parsing log record */
    
     /* scan through log records to find the logs of the undo Tx */
     String tmlsn,                                /* lsn from tm */
           lmlsn,                                 /* lsn from lm */
           clrlsn;
     int size=lsnlist.size();

     ThLogRecord logrec;  /* get the device's name that stores log records */
     String filename = (String)sys.getProperties().get("logRecords");
     int i = 0;
      
     try                                    /* Open the file name */
     {
        inStream = new DataInputStream(
                   new BufferedInputStream(
                    new FileInputStream(filename)));  
        
        /* Search for the log record with the LSN number */
        while ((inline = inStream.readLine())!=null)
        {
           tk = new StringTokenizer(inline);   
           logtype = tk.nextToken();
           lsn = tk.nextToken();
           while (i < size)
           {  
              tmlsn = (String)lsnlist.get(i);        
              if(tmlsn.equals(lsn))
              {
                 logrec = createLogRecord(tk, sys, logtype, lsn);
                 logs.add(0, logrec); 
                 i++;
              }               
              break;
           }
        }
       return logs;
     }
     catch (Exception e)
     {
        e.toString();
        return null;
     }    
  }
}