package cs297.xmldb.src.sm;
import java.io.*;
import java.util.*;
import java.lang.*;
import cs297.xmldb.src.cm.*;
import cs297.xmldb.src.bm.*;
import cs297.xmldb.src.server.*;

/**
 COPYRIGHT (C) 2004 Thien An Nguyen. All Rights Reserved.
 This class handles recovery of system crash.
 @Date: Spring & fall 2004
 @author: Thien An Nguyen, 
 */
public class ThRecoveryMgr implements Serializable
{   
  public ThRecoveryMgr(Properties properties)
  {
  }
  
  /**
   In this phase, recovery manager gets the redo and undo 
   log records lists. It also sort the list and Natix's 
   LSN might be out of order due to private log records.
   @params: the system object, redo list and undo list
   @return: after this function returns, it fills up the 
            redo and undo list of log records.
   */
  private void analyzeLogs(ThSystem sys, ArrayList redologs, ArrayList undologs)
  {
     ThTransactionMgr TM = sys.getTM();
     ThTransaction tx; 
     String txid;
     ArrayList redolsns = new ArrayList(),
               undolsns = new ArrayList();
     sys.out("Analyzing Logs.....");
    
     for (Enumeration e=TM.keys(); e.hasMoreElements();)
     {
        txid = (String) e.nextElement();
        tx = TM.get(txid);
        tx.getLSNsForRedoS(sys.isNatix(), redolsns);
        tx.getLSNsForUndoS(sys.isNatix(), undolsns); 
     }
    
     redologs.addAll(sys.getLM().getLogRecords(redolsns, sys));  
     undologs.addAll(sys.getLM().getLogRecords(undolsns, sys));
     java.util.Collections.sort(redologs);
     java.util.Collections.sort(undologs);
  }
  
  /**
   This function performs the rolls back for uncommitted transaction
   @param: the log records to be undone
           system object
   */
  public void undoS(ArrayList logrecs, ThSystem sys)
  {
     sys.out("\nUndo Phase!.....\n");
     ThLogRecord log = null;
     ThXMLDoc xmldoc = null;
     int size = logrecs.size();
     for (int i=size-1; i>=0; i--)
     {
        log = (ThLogRecord)logrecs.get(i);
        String xmlname = ((ThLogRecord)logrecs.get(i)).getXmlName();
        String dbname = sys.getDB(xmlname);
        ThDB db = (ThDB)sys.getCatalog().get(dbname);
        xmldoc = (ThXMLDoc)db.getDocNames().get(xmlname);
        /*undo actions for un-commited txs */
        xmldoc.undo(log, sys); 
     }         
  }
  
  /**
   This function performs the rolls back for committed transactions
   @param: the log records to be undone
           system object
   */
  private void redoS(ArrayList logrecs, ThSystem sys)
  {
     sys.out("\nRedo Phase!.....\n");
     ThLogRecord log = null;
     ThXMLDoc xmldoc = null;
     int size = logrecs.size();
     for (int i=size-1; i>=0; i--)
     {
        log = (ThLogRecord)logrecs.get(i);
        String xmlname = ((ThLogRecord)logrecs.get(i)).getXmlName();
        String dbname = sys.getDB(xmlname);
        ThDB db = (ThDB)sys.getCatalog().get(dbname);
        xmldoc = (ThXMLDoc)db.getDocNames().get(xmlname);
        /*re-apply for commited txs */
        xmldoc.redo(log, sys); 
     }     
  }
  
  /**
   This function handles for rolls back operations.
   It also writes a CLR for this operations.
   @params: transaction id to be rolls back,
            system object
   @return: return code
   */
  public int rollback(String txid, ThSystem sys)
  {    
      ThTransactionMgr TM = sys.getTM();
      ThTransaction tx = TM.get(txid);     
      sys.out("Rollback!");
    
      /* get log records of the same transaction from log manager */
      ArrayList lsns = new ArrayList();
      String clrlsn= tx.getLSNsForRollback(lsns);
      ArrayList logrecs = sys.getLM().getLogRecords(lsns, sys);
 
      createCompensationLogRecord(sys, txid, clrlsn);
      /* go through each log record to undo the action */
      ThLogRecord log = null;
      ThXMLDoc xmldoc = null;
      for (int i=0; i<logrecs.size(); i++)
      {
        log = (ThLogRecord)logrecs.get(i);
        String xmlname = ((ThLogRecord)logrecs.get(i)).getXmlName();
        String dbname = sys.getDB(xmlname);
        ThDB db = (ThDB)sys.getCatalog().get(dbname);
        xmldoc = (ThXMLDoc)db.getDocNames().get(xmlname);
        xmldoc.undo(log, sys); 
      }
 
    return 0;
  }
  
  /**
   This function performs the three recovery phases: 
   analyzing log records, redo phase, and undo phase.
   @param: system object
   @return return code
   */
  public int restartRecovery(ThSystem sys)
  {
    ArrayList redologs = new ArrayList();
    ArrayList undologs = new ArrayList();
    /* analyze logs */
    analyzeLogs(sys, redologs, undologs);

    /* Redo for commited transactions */
    redoS(redologs, sys);
    
    /* Undo for un-commited transactions */
    undoS(undologs, sys);
    
    return 0;
  }
  
  /**
   This function handle the creation of the compensation log record.
   @param: system object, 
           transaction id,
           the last LSN
   @return: nothing
   */
  public void createCompensationLogRecord(ThSystem sys, 
                              String txid, String lastLSN)
  {
     String lsn = sys.getLSN(); //lsn of this l record
     String data = "CLR " + lsn + " " + lastLSN;
     sys.getLM().writeLogRecord(data);  
  }
}