package cs297.xmldb.src.sm;
import java.io.*;
import java.util.*;
import java.lang.*;
import cs297.xmldb.src.cm.*;
import cs297.xmldb.src.server.*;

/**
  COPYRIGHT (C) 2004 Thien An Nguyen. All Rights Reserved.
  Contains information about a transaction.
  @Date: Spring & fall 2004
  @author: Thien An Nguyen, 
 */
public class ThTransaction
{  
   String iD;
   String LSN;
   boolean committed = false;
   ArrayList lsnRecs; /*chain of log record LSNs of 
                       this transaction */
   
   public ThTransaction(String id)
   {
      iD = id;
      committed = false;
      lsnRecs = new ArrayList(); 
   }
  
   public int addUndoLSN(String lsn)
   {
      lsnRecs.add(lsn);
      committed = false;
    
      return 0;
   }  
  
   public boolean isCommitted() 
   {
      return committed;
   }
  
   public void setCommitted(boolean val) 
   {
      committed = val;
   }
  
   public void savePoint()
   {
      committed = true;
   }  
  
   public void commit()
   {
      lsnRecs.add("COMMIT");
      committed = true;
   }  
  
   public void archive()
   {
      lsnRecs.add("ARCHIVE");
   }  
   
   public void setLSN(String newval) 
   {
      LSN = newval;
   } 
  
   public ArrayList getLSNs() 
   {
      return lsnRecs;
   } 
   
   /**
    If this transaction is not currently in commited
    state, traverse backward through the list of lsn 
    of this transaction until see a committed mark 
    and copy the lsn numbers into the return arraylist.
    If this transaction is already committed, return 
    the arraylist with no elements in it.
    @return list of lsn for roll back
    */
   public String getLSNsForRollback(ArrayList lsnlist)
   {     
      if (committed)
         return "";
    
      String lsn;
      for (int i = lsnRecs.size()-1; i >= 0; i--)
      {
         lsn = (String)lsnRecs.get(i);
         if (lsn.equals("COMMIT"))
         {
            lsn = (String)lsnRecs.get(i-1);
            removeRollbackLSNs();
            return lsn;
         }
         //add lsn into list for roll back
         lsnlist.add(0, lsn);
      }
/*      for (int i = lsnRecs.size()-1; i >= 0; i--)
      {
         lsn = (String)lsnRecs.get(i);
         if (lsn.equals("COMMIT"))
         {
            lsn = (String)lsnRecs.get(i-1);
            lsnlist.add(0, lsn);
            return lsnlist;
         }
      
         lsnlist.add(0, lsn);
      } */
      
      removeFromArray("ARCHIVE", lsnlist);
      return "";
   }  
  
   private void removeRollbackLSNs()
   {
      if (committed)
         return;

      for (int i = lsnRecs.size()-1; i >= 0; i--)
      {
         String lsn = (String)lsnRecs.get(i);
         if (lsn.equals("COMMIT"))
            return;
         else
            lsnRecs.remove(i);
      }       
   }
   
   private void removeFromArray(String key, ArrayList arr)
   {
      while (arr.indexOf(key) >= 0)
         arr.remove(key);
   }
  
   private void removeRange(ArrayList arr, int start, int end)
   {
      for (int i = end-1; i >= start; i--)
         arr.remove(i);
   }
  
   /**
    */
   public int getLSNsForRedoS(boolean isnatix, ArrayList list)
   {       
      int from = lsnRecs.lastIndexOf("ARCHIVE");
      int to   = lsnRecs.size();
      if (!isnatix)
      {
         if (from < 0)
           from = 0;
         for (int i = from+1; i < to; i++)
            if (!((String)lsnRecs.get(i)).equals("COMMIT"))
               list.add(lsnRecs.get(i)); 
         
         return 0;
      }
      
      to   = lsnRecs.lastIndexOf("COMMIT");
      
      if (from >= 0 && to - from <= 1) //nothing to redo
        return 0;
      
      if (from >= 0 && from < to)
      { //could have more than 1 commit
         for (int i = to-1; i > from; i--) 
            if (!((String)lsnRecs.get(i)).equals("COMMIT"))
               list.add(0, lsnRecs.get(i));
      }
      else if (to > 0) 
      { //never archive of database
         for (int i = to-1; i >= 0; i--) 
            if (!((String)lsnRecs.get(i)).equals("COMMIT"))
               list.add(0, lsnRecs.get(i));
      }
      
      return 0; // success
   }    
   
   /**
   */
   public int getLSNsForUndoS(boolean isnatix, ArrayList list)
   {      
      int from = lsnRecs.lastIndexOf("ARCHIVE");
      int to   = lsnRecs.size();
      if (!isnatix)
      {
             if (from < 0)
                from = 0;
             for (int i = to - 1; i > from; i--)
                if (!((String)lsnRecs.get(i)).equals("COMMIT"))
                {
                  list.add(0, lsnRecs.get(i)); 
                  return 0;
                }
         
         return 0;
      } 
      
      from = lsnRecs.lastIndexOf("ARCHIVE");
      to = lsnRecs.lastIndexOf("COMMIT");
      
      if (to >= 0 && from - to <= 1) //nothing to redo
        return 0;
      
      if (to >= 0 && to < from)
         for (int i=from-1; i>to; i--) 
            list.add(0, lsnRecs.get(i));
      else if (to > 0)  //never archive of database
         for (int i=to-1; i>=0; i--) 
            if (!((String)lsnRecs.get(i)).equals("COMMIT"))
               list.add(0, lsnRecs.get(i));

      return 0; // success       
   }
  
   public String getId() 
   {
      return iD;
   }
   
   public String getLastLSN(String currlsn)
   {
      String lastlsn = null;
      int idx = lsnRecs.indexOf(currlsn);
      if (idx > 0)
         lastlsn = (String) lsnRecs.get(idx-1);
     
      return lastlsn;
   }
}